# -*- coding: utf-8 -*-
"""reV power curve losses module.

"""
import json
import logging
import warnings
from abc import ABC, abstractmethod

import numpy as np
from scipy.optimize import minimize_scalar

from reV.utilities.exceptions import reVLossesValueError, reVLossesWarning
from reV.losses.utils import _validate_arrays_not_empty

logger = logging.getLogger(__name__)


class PowerCurve:
    """A turbine power curve.

    Attributes
    ----------
    wind_speed : :obj:`np.array`
        An array containing the wind speeds corresponding to the values
        in the :attr:`power_curve` array.
    generation : :obj:`np.array`
        An array containing the generated power at the corresponding
        wind speed in the :attr:`wind_speed` array. This input must have
        at least one positive value, and if a cutoff speed is detected
        (see `Warnings` section below), then all values above that wind
        speed must be set to 0.

    Warnings
    --------
    This class will attempt to infer a cutoff speed from the
    ``generation`` input. Specifically, it will look for a transition
    from the highest rated power down to zero in a single ``wind_speed``
    step of the power curve. If such a transition is detected, the wind
    speed corresponding to the zero value will be set as the cutoff
    speed, and all calculated power curves will be clipped at this
    speed. If your input power curve contains a cutoff speed, ensure
    that it adheres to the expected pattern of dropping from max rated
    power to zero power in a single wind speed step.
    """
    def __init__(self, wind_speed, generation):
        """
        Parameters
        ----------
        wind_speed : iter
            An iterable containing the wind speeds corresponding to the
            generated power values in ``generation`` input. The input
            values should all be non-zero.
        generation : iter
            An iterable containing the generated power at the
            corresponding wind speed in the ``wind_speed`` input. This
            input must have at least one positive value, and if a cutoff
            speed is detected (see `Warnings` section below), then all
            values above that wind speed must be set to 0.
        """
        self.wind_speed = np.array(wind_speed)
        self.generation = np.array(generation)
        self._cutoff_wind_speed = None

        _validate_arrays_not_empty(self,
                                   array_names=['wind_speed', 'generation'])
        self._validate_wind_speed()
        self._validate_generation()

    def _validate_wind_speed(self):
        """Validate that the input wind speed is non-negative. """
        if not (self.wind_speed >= 0).all():
            msg = ("Invalid wind speed input: Contains negative values! - {}"
                   .format(self.wind_speed))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_generation(self):
        """Validate the input generation. """
        if not (self.generation > 0).any():
            msg = ("Invalid generation input: Found no positive values! - {}"
                   .format(self.generation))
            logger.error(msg)
            raise reVLossesValueError(msg)

        if 0 < self.cutoff_wind_speed < np.inf:
            cutoff_windspeed_ind = np.where(self.wind_speed
                                            >= self.cutoff_wind_speed)[0].min()
            if (self.generation[cutoff_windspeed_ind:]).any():
                msg = ("Invalid generation input: Found non-zero values above "
                       "cutoff! - {}".format(self.generation))
                logger.error(msg)
                raise reVLossesValueError(msg)

    @property
    def cutoff_wind_speed(self):
        """The detected cutoff wind speed at which the power generation is zero

        Returns
        --------
        float | np.inf
        """
        if self._cutoff_wind_speed is None:
            ind = np.argmax(self.generation[::-1])
            # pylint: disable=chained-comparison
            if ind > 0 and self.generation[-ind] <= 0:
                self._cutoff_wind_speed = self.wind_speed[-ind]
            else:
                self._cutoff_wind_speed = np.inf
        return self._cutoff_wind_speed

    def __eq__(self, other):
        return np.isclose(self.generation, other).all()

    def __ne__(self, other):
        return not np.isclose(self.generation, other).all()

    def __lt__(self, other):
        return self.generation < other

    def __le__(self, other):
        return self.generation <= other

    def __gt__(self, other):
        return self.generation > other

    def __ge__(self, other):
        return self.generation >= other

    def __len__(self):
        return len(self.generation)

    def __getitem__(self, index):
        return self.generation[index]

    def __call__(self, wind_speed):
        """Calculate the power curve value for the given ``wind_speed``.

        Parameters
        ----------
        wind_speed : int | float | list | np.array
            Wind speed value corresponding to the desired powerrcurve
            value.

        Returns
        -------
        float | :obj:`np.array`
            The power curve value(s) for the input wind speed(s).
        """
        if isinstance(wind_speed, (int, float)):
            wind_speed = np.array([wind_speed])
        power_generated = np.interp(wind_speed, self.wind_speed,
                                    self.generation)
        if self.cutoff_wind_speed:
            power_generated[wind_speed >= self.cutoff_wind_speed] = 0
        return power_generated


class PowerCurveLosses:
    """A converter between annual losses and power curve transformation.

    Given a target annual loss value, this class facilitates the
    calculation of a power curve transformation such that the annual
    generation losses incurred by using the transformed power curve when
    compared to the original (non-transformed) power curve match the
    target loss as close as possible.

    The underlying assumption for this approach is that some types of
    losses can be realized by a transformation of the power curve (see
    the values of :obj:`TRANSFORMATIONS` for details on all of the
    power curve transformations that have been implemented).

    The advantage of this approach is that, unlike haircut losses (where
    a single loss value is applied across the board to all generation),
    the losses are distributed non-uniformly across the power curve. For
    example, even in the overly simplified case of a horizontal
    translation of the power curve (which is only physically realistic
    for certain types of losses like blade degradation), the losses are
    distributed primarily across region 2 of the power curve (the steep,
    almost linear, portion where the generation rapidly increases). This
    means that, unlike with haircut losses, generation is able to reach
    max rated power (albeit at a greater wind speed).

    Attributes
    ----------
    power_curve : :obj:`PowerCurve`
        A :obj:`PowerCurve` object representing the "original
        powerCurve".
    wind_resource : :obj:`np.array`
        An array containing the wind speeds (i.e. wind speed
        distribution) for the site at which the power curve will be
        used. This distribution is used to calculate the annual
        generation of the original power curve as well as any additional
        calcaulted power curves. The generation values are then compared
        in order to calculate the loss resulting from a transformed
        power curve.
    """

    def __init__(self, power_curve, wind_resource):
        """
        Parameters
        ----------
        power_curve : PowerCurve
            A :obj:`PowerCurve` object representing the turbine
            power curve. This input is treated as the
            "original" power curve.
        wind_resource : iter
            An iterable containing the wind speeds measured at the site
            where this power curve will be applied to caulcate
            generation. These values are used to calculate the loss
            resulting from a transformed power curve compared to the
            generation of the original power curve. The input
            values should all be non-zero, and the units of
            should match the units of the ``power_curve`` input
            (typically, m/s).
        """
        self.power_curve = power_curve
        self.wind_resource = np.array(wind_resource)
        self._power_gen = None

        _validate_arrays_not_empty(self, array_names=['wind_resource'])
        self._validate_wind_resource()

    def _validate_wind_resource(self):
        """Validate that the input wind resource is non-negative. """
        if not (self.wind_resource >= 0).all():
            msg = ("Invalid wind resource input: Contains negative values!"
                   " - {}".format(self.wind_resource))
            msg = msg.format(self.wind_resource)
            logger.error(msg)
            raise reVLossesValueError(msg)

    def annual_losses_with_transformed_power_curve(
        self, transformed_power_curve
    ):
        """Calculate the annual losses from a transformed power curve.

        This function uses the wind resource data that the object was
        initialized with to calculate the total annual power generation
        with a transformed power curve. This generation is compared with
        the generation of the original (non-transformed) power curve to
        compute the total annual losses as a result of the
        transformation.

        Parameters
        ----------
        transformed_power_curve : PowerCurve
            A :obj:`PowerCurve` object representing the transformed
            power curve. The power generated with this power curve will
            be compared with the power generated by the "original" power
            curve to calculate annual losses.

        Returns
        -------
        float
            Total losses (%) as a result of a the power curve
            transformation.
        """
        power_gen_with_losses = transformed_power_curve(self.wind_resource)
        power_gen_with_losses = power_gen_with_losses.sum()
        return (1 - power_gen_with_losses / self.power_gen_no_losses) * 100

    def _obj(self, transformation_variable, target, transformation):
        """Objective function: |output - target|."""
        new_power_curve = transformation.apply(transformation_variable)
        losses = self.annual_losses_with_transformed_power_curve(
            new_power_curve
        )
        return np.abs(losses - target)

    def fit(self, target, transformation):
        """Fit a power curve transformation.

        This function fits a transformation to the input power curve
        (the one used to initialize the object) to generate an annual
        loss percentage closest to the ``target``. The losses are
        computed w.r.t the generation of the original (non-transformed)
        power curve.

        Parameters
        ----------
        target : float
            Target value for annual generation losses (%).
        transformation : PowerCurveTransformation
            A :class:`PowerCurveTransformation` class representing the
            power curve transformation to use.

        Returns
        -------
        :obj:`np.array`
            An array containing a transformed power curve that most
            closely yields the ``target`` annual generation losses.

        Warns
        -----
        reVLossesWarning
            If the fit did not meet the target annual losses to within
            1%.

        Warnings
        --------
        This function attempts to find an optimal transformation for the
        power curve such that the annual generation losses match the
        ``target`` value, but there is no guarantee that a close match
        can be found, if it even exists. Therefore, it is possible that
        the losses resulting from the transformed power curve will not
        match the ``target``. This is especially likely if the
        ``target`` is large or if the input power curve and/or wind
        resource is abnormal.
        """
        transformation = transformation(self.power_curve)
        fit_var = minimize_scalar(self._obj,
                                  args=(target, transformation),
                                  bounds=transformation.bounds,
                                  method='bounded').x
        loss_percentage = self._obj(fit_var, target, transformation)
        if loss_percentage > 1:
            msg = ("Unable to find a transformation such that the losses meet "
                   "the target within 1%! Obtained fit with loss percentage "
                   "{}% when target was {}%. Consider using a different "
                   "transformation or reducing the target losses!"
                   .format(loss_percentage, target))
            logger.warning(msg)
            warnings.warn(msg, reVLossesWarning)
        return transformation.apply(fit_var)

    @property
    def power_gen_no_losses(self):
        """float: Total power generation from original power curve."""
        if self._power_gen is None:
            self._power_gen = self.power_curve(self.wind_resource).sum()
        return self._power_gen


class PowerCurveLossesInput:
    """Power curve losses specification.

    This class stores and validates information about the desired losses
    from a given type of power curve transfromation. In particular, the
    target loss percentage must be provided. This input is then
    validated to be used powercurve transformation fitting.

    """

    REQUIRED_KEYS = {'target_losses_percent'}
    """Required keys in the input specification dictionary."""

    def __init__(self, specs):
        """

        Parameters
        ----------
        specs : dict
            A dictionary containing specifications for the power curve
            losses. This dictionary must contain the following keys:
                - `target_losses_percent`
                    An integer or float value representing the
                    total percentage of annual energy production that
                    should be lost due to the powercurve transformation.
                    This value must be in the range [0, 100].
            The input dictionary can also provide the following optional
            keys:
                - `transformation` - by default, ``horizontal_translation``
                    A string representing the type of transformation to
                    apply to the power curve. This sting must be one of
                    the keys of :obj:`TRANSFORMATIONS`.

        """
        self._specs = specs
        self._transformation_name = self._specs.get('transformation',
                                                    'exponential_stretching')
        self._validate()

    def _validate(self):
        """Validate the input specs."""
        self._validate_required_keys_exist()
        self._validate_transformation()
        self._validate_percentage()

    def _validate_required_keys_exist(self):
        """Raise an error if any required keys are missing."""
        missing_keys = [n not in self._specs for n in self.REQUIRED_KEYS]
        if any(missing_keys):
            msg = ("The following required keys are missing from the power "
                   "curve losses specification: {}"
                   .format(sorted(missing_keys)))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_transformation(self):
        """Validate that the transformation exists in TRANSFORMATIONS. """
        if self._transformation_name not in TRANSFORMATIONS:
            msg = ("Transformation {!r} not understood! "
                   "Input must be one of: {} "
                   .format(self._transformation_name,
                           list(TRANSFORMATIONS.keys())))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_percentage(self):
        """Validate that the percentage is in the range [0, 100]. """
        if not 0 <= self.target <= 100:
            msg = ("Percentage of annual energy production loss to be "
                   "attributed to the powercurve transformation must be in "
                   "the range [0, 100], but got {} for transformation {!r}"
                   .format(self.target, self._transformation_name))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def __repr__(self):
        specs = self._specs.copy()
        specs.update({'transformation': self._transformation_name})
        specs_as_str = ", ".join(["{}={!r}".format(k, v)
                                  for k, v in specs.items()])
        return "PowerCurveLossesInput({})".format(specs_as_str)

    @property
    def target(self):
        """int or float: Target loss percentage due to transformation."""
        return self._specs['target_losses_percent']

    @property
    def transformation(self):
        """PowerCurveTransformation: Requested power curve transformation."""
        return TRANSFORMATIONS[self._transformation_name]


class PowerCurveLossesMixin:
    """Mixin class for :class:`reV.SAM.generation.AbstractSamWind`.

    Warning
    -------
    Using this class for anything excpet as a mixin for
    :class:`~reV.SAM.generation.AbstractSamWind` may result in unexpected
    results and/or errors.
    """

    POWERCURVE_CONFIG_KEY = 'reV_power_curve_losses'
    """Specify power curve loss target in the config file using this key."""

    def add_power_curve_losses(self):
        """Adjust power curve in SAM config file to account for losses.

        This function reads the information in the
        ``reV_power_curve_losses`` key of the ``sam_sys_inputs``
        dictionary and computes a new power curve that accounts for the
        loss percentage specified from that input. If no power curve
        loss info is specified in ``sam_sys_inputs``, the power curve
        will not be adjusted.

        See Also
        --------
        :class:`PowerCurveLosses` : Power curve re-calculation.

        """
        loss_input = self._user_power_curve_input()
        if not loss_input:
            return

        wind_resource = self.wind_resource_for_site()
        pc_losses = PowerCurveLosses(self.input_power_curve, wind_resource)

        logger.debug("Transforming power curve using the {} transformation to "
                     "meet {}% loss target..."
                     .format(loss_input.transformation, loss_input.target))

        new_curve = pc_losses.fit(loss_input.target, loss_input.transformation)
        logger.debug("Transformed power curve: {}".format(list(new_curve)))
        self.sam_sys_inputs['wind_turbine_powercurve_powerout'] = new_curve

    def _user_power_curve_input(self):
        """Get power curve loss info from config. """
        power_curve_losses_info = self.sam_sys_inputs.pop(
            self.POWERCURVE_CONFIG_KEY, None
        )
        if power_curve_losses_info is None:
            return

        # site-specific info is input as str
        if isinstance(power_curve_losses_info, str):
            power_curve_losses_info = json.loads(power_curve_losses_info)

        loss_input = PowerCurveLossesInput(power_curve_losses_info)
        if loss_input.target <= 0:
            logger.debug("Power curve target loss is 0. Skipping powercurve "
                         "transformation.")
            return

        return loss_input

    @property
    def input_power_curve(self):
        """PowerCurve: Original power curve for site. """
        wind_speed = self.sam_sys_inputs['wind_turbine_powercurve_windspeeds']
        generation = self.sam_sys_inputs['wind_turbine_powercurve_powerout']
        return PowerCurve(wind_speed, generation)

    def wind_resource_for_site(self):
        """Extract scaled wind speeds at the site.

        Get the wind resource (wind speeds) for this site, accounting
        for the scaling done in SAM based on air pressure. These wind
        speeds can then be used to sample the power curve and obtain
        generation values.

        Returns
        -------
        array-like
            Array of scaled wind speeds.

        References
        ----------
        Scaling done in SAM:
            https://tinyurl.com/2uzjawpe
        SAM Wind Power Reference Manual for explanations on generation
        and air density calculations (pp. 18):
            https://tinyurl.com/2p8fjba6

        """

        temperatures, pressures, wind_speeds, __, = map(
            np.array, zip(*self['wind_resource_data']['data'])
        )

        pressures_pascal = pressures * 101325.0273830  # originally in atm
        temperatures_K = temperatures + 273.15  # originally in celsius
        specific_gas_constant_dry_air = 287.058  # units: J / kg / K
        sea_level_air_density = 1.225  # units: kg/m**3 at 15 degrees celsius

        site_air_densities = pressures_pascal / (specific_gas_constant_dry_air
                                                 * temperatures_K)
        weights = (sea_level_air_density / site_air_densities) ** (1 / 3)
        return wind_speeds / weights


class AbstractPowerCurveTransformation(ABC):
    """Abstract base class for power curve transformations.

    **This class is not meant to be instantiated**.

    This class provides an interface for power curve transformations,
    which are meant to more realistically represent certain types of
    losses when compared to simple haircut losses (i.e. constant loss
    value applied at all points on the power curve).

    If you would like to implement your own power curve transfromation,
    you should subclass this class and implement the :meth:`apply`
    method and the :attr:`bounds` property. See the documentation for
    each of these below for more details.

    Attributes
    ----------
    power_curve : :obj:`PowerCurve`
        A :obj:`PowerCurve` object representing the "original" power
        curve.
    """
    def __init__(self, power_curve):
        """
        Parameters
        ----------
        power_curve : PowerCurve
            A :obj:`PowerCurve` object representing the turbine
            power curve. This input is treated as the "original" power
            curve.
        """
        self.power_curve = power_curve
        self._transformed_generation = None

    def _validate_shifted_power_curve(self, new_curve):
        """Ensure new power curve has some non-zero generation. """
        mask = (self.power_curve.wind_speed
                <= self.power_curve.cutoff_wind_speed)
        min_expected_power_gen = self.power_curve[self.power_curve > 0].min()
        if not (new_curve[mask] > min_expected_power_gen).any():
            msg = ("Calculated power curve is invalid. No power generation "
                   "below the cutoff wind speed ({} m/s) detected. Target "
                   "loss percentage  may be too large! Please try again with "
                   "a lower target value."
                   .format(self.power_curve.cutoff_wind_speed))
            logger.error(msg)
            raise reVLossesValueError(msg)

    @abstractmethod
    def apply(self, transformation_var):
        """Apply a transformation to the original power curve.

        Parameters
        ----------
        transformation_var : : float
            A single variable controlling the "strength" of the
            transformation. The :class:`PowerCurveLosses` object will
            run an optimization using this variable to fit the target
            annual losses incurred with the transformed power curve
            compared to the original power curve using the given wind
            resource distribution.

        Returns
        -------
        :obj:`PowerCurve`
            An new power curve containing the generation values from the
            transformed power curve.

        Notes
        -----
        When implementing a new transformation, override this method and
        set the ``_transformed_generation`` protected attribute to be
        the generation corresponding to the transformed power curve.
        Then, call ``super().apply(transformation_var)`` in order to
        apply cutout speed curtailment and validation for the
        transformed power curve. For example, here is the implementation
        for a transformation that shifts the power curve horizontally::

            self._transformed_generation = self.power_curve(
                self.power_curve.wind_speed - transformation_var
            )
            return super().apply(transformation_var)

        """
        if self._transformed_generation is None:
            msg = ("Transformation implementation for {}.apply did not set "
                   "the `_transformed_generation` attribute."
                   .format(type(self).__name__))
            logger.error(msg)
            raise NotImplementedError(msg)

        mask = (self.power_curve.wind_speed
                >= self.power_curve.cutoff_wind_speed)
        self._transformed_generation[mask] = 0

        new_curve = PowerCurve(self.power_curve.wind_speed,
                               self._transformed_generation)
        self._validate_shifted_power_curve(new_curve)
        return new_curve

    @property
    @abstractmethod
    def bounds(self):
        """tuple: Bounds on the ``transformation_var``."""


class HorizontalTranslation(AbstractPowerCurveTransformation):
    """Utility for applying horizontal power curve translations.

    The mathematical representation of this transformation is:

    .. math:: P_{transformed}(u) = P_{original}(u - t),

    where :math:`P_{transformed}` is the transformed power curve,
    :math:`P_{original}` is the original power curve, :math:`u` is
    the wind speed, and :math:`t` is the transformation variable
    (horizontal translation amount).

    This kind of power curve transformation is simplistic, and should
    only be used for a small handful of applicable turbine losses
    (i.e. blade degradation). See ``Warnings`` for more details.

    The losses in this type of transformation are distributed primarily
    across region 2 of the power curve (the steep, almost linear,
    portion where the generation rapidly increases):

    .. image:: ../../../examples/rev_losses/horizontal_translation.png
       :align: center

    Attributes
    ----------
    power_curve : :obj:`PowerCurve`
        A :obj:`PowerCurve` object representing the "original" power
        curve.

    Warnings
    --------
    This kind of power curve translation is not genrally realistic.
    Using this transformation as a primary source of losses (i.e. many
    different kinds of losses bundled together) is extremely likely to
    yield unrealistic results!
    """

    def apply(self, transformation_var):
        """Apply a horizontal translation to the original power curve.

        This function shifts the original power curve horizontally,
        along the "wind speed" (x) axis, by the given amount. Any power
        above the cutoff speed (if one was detected) is truncated after
        the transforamtion.

        Parameters
        ----------
        transformation_var : float
            The amount to shift the original power curve by, in wind
            speed units (typically, m/s).

        Returns
        -------
        :obj:`PowerCurve`
            An new power curve containing the generation values from the
            shifted power curve.
        """
        self._transformed_generation = self.power_curve(
            self.power_curve.wind_speed - transformation_var
        )
        return super().apply(transformation_var)

    @property
    def bounds(self):
        """tuple: Bounds on the power curve shift for the fitting procedure."""
        min_ind = np.where(self.power_curve)[0][0]
        max_ind = np.where(self.power_curve[::-1])[0][0]
        max_shift = (self.power_curve.wind_speed[-max_ind - 1]
                     - self.power_curve.wind_speed[min_ind])
        return (0, max_shift)


class LinearStretching(AbstractPowerCurveTransformation):
    """Utility for applying a linear stretch to the power curve.

    The mathematical representation of this transformation is:

    .. math:: P_{transformed}(u) = P_{original}(u/t),

    where :math:`P_{transformed}` is the transformed power curve,
    :math:`P_{original}` is the original power curve, :math:`u` is
    the wind speed, and :math:`t` is the transformation variable
    (wind speed multiplier).

    The losses in this type of transformation are distributed primarily
    across regions 2 and 3 of the power curve. In particular, losses are
    smaller for wind speeds closer to the cut-in speed, and larger for
    speeds close to rated power:

    .. image:: ../../../examples/rev_losses/linear_stretching.png
       :align: center

    Attributes
    ----------
    power_curve : :obj:`PowerCurve`
        A :obj:`PowerCurve` object representing the "original" power
        curve.
    """

    def apply(self, transformation_var):
        """Apply a linear stretch to the original power curve.

        This function stretches the original power curve along the
        "wind speed" (x) axis. Any power above the cutoff speed (if one
        was detected) is truncated after the transforamtion.

        Parameters
        ----------
        transformation_var : float
            The linear multiplier of the wind speed scaling.

        Returns
        -------
        :obj:`PowerCurve`
            An new power curve containing the generation values from the
            shifted power curve.
        """
        self._transformed_generation = self.power_curve(
            self.power_curve.wind_speed / transformation_var
        )
        return super().apply(transformation_var)

    @property
    def bounds(self):
        """tuple: Bounds on the wind speed multiplier."""
        min_ind_pc = np.where(self.power_curve)[0][0]
        min_ind_ws = np.where(self.power_curve.wind_speed > 1)[0][0]
        min_ws = self.power_curve.wind_speed[max(min_ind_pc, min_ind_ws)]
        max_ws = min(self.power_curve.wind_speed.max(),
                     self.power_curve.cutoff_wind_speed)
        max_multiplier = np.ceil(max_ws / min_ws)
        return (1, max_multiplier)


class ExponentialStretching(AbstractPowerCurveTransformation):
    """Utility for applying an exponential stretch to the power curve.

    The mathematical representation of this transformation is:

    .. math:: P_{transformed}(u) = P_{original}(u^{1/t}),

    where :math:`P_{transformed}` is the transformed power curve,
    :math:`P_{original}` is the original power curve, :math:`u` is
    the wind speed, and :math:`t` is the transformation variable
    (wind speed exponent).

    The losses in this type of transformation are distributed primarily
    across regions 2 and 3 of the power curve. In particular, losses are
    smaller for wind speeds closer to the cut-in speed, and larger for
    speeds close to rated power:

    .. image:: ../../../examples/rev_losses/exponential_stretching.png
       :align: center

    Attributes
    ----------
    power_curve : :obj:`PowerCurve`
        A :obj:`PowerCurve` object representing the "original" power
        curve.
    """

    def apply(self, transformation_var):
        """Apply an exponential stretch to the original power curve.

        This function stretches the original power curve along the
        "wind speed" (x) axis. Any power above the cutoff speed (if one
        was detected) is truncated after the transforamtion.

        Parameters
        ----------
        transformation_var : float
            The exponent of the wind speed scaling.

        Returns
        -------
        :obj:`PowerCurve`
            An new power curve containing the generation values from the
            shifted power curve.
        """
        self._transformed_generation = self.power_curve(
            self.power_curve.wind_speed ** (1 / transformation_var)
        )
        return super().apply(transformation_var)

    @property
    def bounds(self):
        """tuple: Bounds on the wind speed exponent."""
        min_ind_pc = np.where(self.power_curve)[0][0]
        min_ind_ws = np.where(self.power_curve.wind_speed > 1)[0][0]
        min_ws = self.power_curve.wind_speed[max(min_ind_pc, min_ind_ws)]
        max_ws = min(self.power_curve.wind_speed.max(),
                     self.power_curve.cutoff_wind_speed)
        max_exponent = np.ceil(np.log(max_ws) / np.log(min_ws))
        return (1, max_exponent)


TRANSFORMATIONS = {
    'horizontal_translation': HorizontalTranslation,
    'linear_stretching': LinearStretching,
    'exponential_stretching': ExponentialStretching
}
"""Implemented power curve transformations."""



# -*- coding: utf-8 -*-
"""
reV job status manager.
"""
import copy
import os
import json
import logging
import time
from warnings import warn
import shutil

from rex.utilities import safe_json_load
from rex.utilities.hpc import SLURM
from rex.utilities.execution import SubprocessManager

logger = logging.getLogger(__name__)


class Status(dict):
    """Base class for reV data pipeline health and status information."""

    FROZEN_STATUS = ('successful', 'failed')

    def __init__(self, status_dir, name=None, hardware='slurm',
                 subprocess_manager=None):
        """
        Parameters
        ----------
        status_dir : str
            Directory to with json status file.
        name : str | None
            Optional job name for status. Will look for the file
            "{name}_status.json" in the status_dir.
        hardware : str
            Name of hardware that this pipeline is being run on: eagle, slurm,
            local. Defaults to "slurm". This specifies how job are queried for
            status.
        subprocess_manager : None | SLURM
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC.
        """

        self._subprocess_manager = subprocess_manager
        self._hardware = hardware.lower()
        self._status_dir = status_dir
        self._fpath = self._parse_fpath(status_dir, name)
        self.data = self._load(self._fpath)

    @staticmethod
    def _parse_fpath(status_dir, name):
        """Get the status filepath from the status directory and jobname.

        Parameters
        ----------
        status_dir : str
            Directory to with json status file.
        name : str | None
            Optional job name for status. Will look for the file
            "{name}_status.json" in the status_dir.

        Returns
        -------
        fpath : str
            Filepath to job status json.
        """

        if str(status_dir).endswith('.json'):
            raise TypeError('Need a directory containing a status json, '
                            'not a status json: {}'.format(status_dir))

        if name is None:
            fpath = os.path.join(status_dir, 'rev_status.json')
            for fn in os.listdir(status_dir):
                if fn.endswith('_status.json'):
                    fpath = os.path.join(status_dir, fn)
                    break
        else:
            fpath = os.path.join(status_dir, '{}_status.json'.format(name))

        return fpath

    @staticmethod
    def _load(fpath):
        """Load status json.

        Parameters
        -------
        fpath : str
            Filepath to job status json.

        Returns
        -------
        data : dict
            JSON file contents loaded as a python dictionary.
        """
        if os.path.isfile(fpath):
            data = safe_json_load(fpath)
        else:
            data = {}
        return data

    def _dump(self):
        """Dump status json w/ backup file in case process gets killed."""

        if not os.path.exists(os.path.dirname(self._fpath)):
            os.makedirs(os.path.dirname(self._fpath))

        backup = self._fpath.replace('.json', '_backup.json')
        self._sort_by_index()
        if os.path.exists(self._fpath):
            shutil.copy(self._fpath, backup)
        with open(self._fpath, 'w') as f:
            json.dump(self.data, f, indent=4, separators=(',', ': '))
        if os.path.exists(backup):
            os.remove(backup)

    def _sort_by_index(self):
        """Sort modules in data dictionary by pipeline index."""

        sortable = True

        for value in self.data.values():
            if 'pipeline_index' not in value:
                sortable = False
                break

        if sortable:
            sorted_keys = sorted(self.data, key=lambda x:
                                 self.data[x]['pipeline_index'])
            self.data = {k: self.data[k] for k in sorted_keys}

    def _get_job_status(self, job_id):
        """Get the job status using pre-defined hardware-specific methods.

        Parameters
        ----------
        job_id : str | int
            SLURM or PBS job submission id.

        Returns
        -------
        status : str | None
            Job status from qstat/squeue. None if no job found.
        """
        status = None
        options = {'eagle': self.subprocess_manager.check_status,
                   'slurm': self.subprocess_manager.check_status,
                   'local': None}
        if job_id:
            try:
                method = options[self.hardware]
            except KeyError as e:
                msg = ('Could not check job on the requested hardware: '
                       '"{}", available options are: {}.'
                       .format(self.hardware, list(options.keys())))
                logger.error(msg)
                raise KeyError(msg) from e
            if method is None:
                status = None
            else:
                status = method(job_id=job_id)

        return status

    def _check_all_job_files(self, status_dir):
        """Look for all single-job job status files in the target status_dir
        and update status.

        Parameters
        ----------
        status_dir : str
            Directory to look for completion file.
        """

        for fname in os.listdir(status_dir):
            if fname.startswith('jobstatus_') and fname.endswith('.json'):
                # wait one second to make sure file is finished being written
                time.sleep(0.01)
                status = safe_json_load(os.path.join(status_dir, fname))
                self.data = self.update_dict(self.data, status)
                os.remove(os.path.join(status_dir, fname))

    @staticmethod
    def _check_job_file(status_dir, job_name):
        """Look for a single-job job status file in the target status_dir.

        Parameters
        ----------
        status_dir : str
            Directory to look for completion file.
        job_name : str
            Job name.

        Returns
        -------
        status : dict | None
            Job status dictionary if completion file found.
        """
        status = None
        target_fname = 'jobstatus_{}.json'.format(job_name)
        for fname in os.listdir(status_dir):
            if fname == target_fname:
                # wait one second to make sure file is finished being written
                time.sleep(0.01)
                status = safe_json_load(os.path.join(status_dir, fname))
                os.remove(os.path.join(status_dir, fname))
                break

        return status

    def _update_job_status(self, module, job_name):
        """Update HPC job and respective job status to the status obj instance.

        Parameters
        ----------
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        """

        # look for completion file.
        current = self._check_job_file(self._status_dir, job_name)

        # Update status data dict and file if job file was found
        if current is not None:
            self.data = self.update_dict(self.data, current)
            self._dump()

        # check job status via hardware if job file not found.
        elif module in self.data:
            # job exists
            if job_name in self.data[module]:

                # init defaults in case job/module not in status file yet
                previous = self.data[module][job_name].get('job_status', None)
                job_id = self.data[module][job_name].get('job_id', None)

                # get job status from hardware
                current = self._get_job_status(job_id)

                # No current status and job was not successful: failed!
                if current is None and previous != 'successful':
                    self.data[module][job_name]['job_status'] = 'failed'

                # do not overwrite a successful or failed job status.
                elif (current != previous
                      and previous not in self.FROZEN_STATUS):
                    self.data[module][job_name]['job_status'] = current

            # job does not yet exist
            else:
                self.data[module][job_name] = {}

    @property
    def subprocess_manager(self):
        """Get the subprocess manager object based on the hardware spec."""

        if self._subprocess_manager is None:
            if self._hardware in ('eagle', 'slurm'):
                self._subprocess_manager = SLURM()
            elif self._hardware == 'local':
                self._subprocess_manager = SubprocessManager
            else:
                msg = ('Cannot recognize requested hardware: {}'
                       .format(self._hardware))
                logger.error(msg)
                raise ValueError(msg)

        return self._subprocess_manager

    def _set_job_status(self, module, job_name, status):
        """Set an updated job status to the object instance.

        Parameters
        ----------
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        status : str
            Status string to set. Must be a status string in
            the FROZEN_STATUS class attribute.
        """

        if module not in self.data:
            raise KeyError('reV pipeline status has not been initialized '
                           'for "{}".'.format(module))
        if job_name not in self.data[module]:
            raise KeyError('reV pipeline status has not been initialized '
                           'for "{}: {}".'.format(module, job_name))

        self.data[module][job_name]['job_status'] = status

    @classmethod
    def _get_attr_list(cls, inp, key='job_id'):
        """Get all job attribute values from the status data dict.

        Parameters
        ----------
        inp : dict
            Job status dictionary.
        key : str
            Key to get values for.

        Returns
        -------
        out : list
            List of values corresponding to the input key for all jobs in inp.
        """

        out = []

        if isinstance(inp, dict):
            if key in inp:
                out = inp[key]
            else:
                for v in inp.values():
                    temp = cls._get_attr_list(v, key=key)

                    if isinstance(temp, list):
                        if any(temp):
                            out += temp
                    elif isinstance(temp, (int, str)):
                        out.append(temp)
        return out

    @property
    def job_ids(self):
        """Get list of job ids."""
        return self._get_attr_list(self.data, key='job_id')

    @property
    def hardware(self):
        """Get the hardware for this pipeline."""
        return self._hardware

    @classmethod
    def update_dict(cls, d, u):
        """Update a dictionary recursively.

        Parameters
        ----------
        d : dict
            Base dictionary to update.
        u : dict
            New dictionary with data to add to d.

        Returns
        -------
        d : dict
            d with data updated from u.
        """

        d = copy.deepcopy(d)

        for k, v in u.items():
            if isinstance(v, dict):
                d[k] = cls.update_dict(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    @staticmethod
    def make_job_file(status_dir, module, job_name, attrs):
        """Make a json file recording the status of a single job.

        Parameters
        ----------
        status_dir : str
            Directory to put json status file.
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        attrs : str
            Dictionary of job attributes that represent the job status
            attributes.
        """
        if job_name.endswith('.h5'):
            job_name = job_name.replace('.h5', '')
        status = {module: {job_name: attrs}}
        fpath = os.path.join(status_dir, 'jobstatus_{}.json'.format(job_name))
        with open(fpath, 'w') as f:
            json.dump(status, f, sort_keys=True, indent=4,
                      separators=(',', ': '))

    @classmethod
    def add_job(cls, status_dir, module, job_name, replace=False,
                job_attrs=None):
        """Add a job to status json.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        replace : bool
            Flag to force replacement of pre-existing job status.
        job_attrs : dict
            Job attributes. Should include 'job_id' if running on HPC.
        """
        if job_name.endswith('.h5'):
            job_name = job_name.replace('.h5', '')

        obj = cls(status_dir)

        if job_attrs is None:
            job_attrs = {}

        if 'hardware' in job_attrs:
            if job_attrs['hardware'] in ('eagle', 'peregrine', 'slurm', 'pbs'):
                if 'job_id' not in job_attrs:
                    msg = ('Key "job_id" should be in kwargs for "{}" if '
                           'adding job from an HPC node.'
                           .format(job_name))
                    logger.warning(msg)
                    warn(msg)

        # check to see if job exists yet
        exists = obj.job_exists(status_dir, job_name, module_name=module)

        # job exists and user has requested forced replacement
        if replace and exists:
            del obj.data[module][job_name]

        # new job attribute data will be written if either:
        #  A) the user requested forced replacement or
        #  B) if the job does not exist
        if replace or not exists:

            if module not in obj.data:
                obj.data[module] = {job_name: job_attrs}
            else:
                obj.data[module][job_name] = job_attrs

            if 'job_status' not in job_attrs:
                obj.data[module][job_name]['job_status'] = 'submitted'

            obj._dump()

    @classmethod
    def job_exists(cls, status_dir, job_name, module_name=None):
        """Check whether a job exists and return a bool.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        job_name : str
            Unique job name identification.
        module_name : str
            reV module that the job belongs to.

        Returns
        -------
        exists : bool
            True if the job exists in the status json.
        """
        if job_name.endswith('.h5'):
            job_name = job_name.replace('.h5', '')

        obj = cls(status_dir)
        exists = False
        if obj.data:
            module_jobs = obj.data.values()
            if module_name is not None:
                module_jobs = [obj.data[module_name]]
            for jobs in module_jobs:
                if jobs:
                    for name in jobs.keys():
                        if name == job_name:
                            exists = True
                            break

        return exists

    @classmethod
    def retrieve_job_status(cls, status_dir, module, job_name,
                            hardware='slurm', subprocess_manager=None):
        """Update and retrieve job status.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        hardware : str
            Name of hardware that this pipeline is being run on: eagle, slurm,
            local. Defaults to "slurm". This specifies how job are queried for
            status.
        subprocess_manager : None | SLURM
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC.

        Returns
        -------
        status : str | None
            Status string or None if job/module not found.
        """
        if job_name.endswith('.h5'):
            job_name = job_name.replace('.h5', '')

        obj = cls(status_dir, hardware=hardware,
                  subprocess_manager=subprocess_manager)
        obj._update_job_status(module, job_name)

        try:
            status = obj.data[module][job_name].get('job_status', None)
        except KeyError:
            status = None

        return status

    @classmethod
    def set_job_status(cls, status_dir, module, job_name, status):
        """Force set a job status to a frozen status and save to status file.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        module : str
            reV module that the job belongs to.
        job_name : str
            Unique job name identification.
        status : str
            Status string to set. Must be a status string in
            the FROZEN_STATUS class attribute.
        """
        if job_name.endswith('.h5'):
            job_name = job_name.replace('.h5', '')

        obj = cls(status_dir)
        obj._set_job_status(module, job_name, status)
        obj._dump()

    @classmethod
    def update(cls, status_dir):
        """Update all job statuses and dump to json.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        """

        obj = cls(status_dir)
        for module in obj.data.keys():
            for job_name in obj.data[module].keys():
                if job_name != 'pipeline_index':
                    obj._update_job_status(module, job_name)
        obj._check_all_job_files(status_dir)
        obj._dump()



# -*- coding: utf-8 -*-
"""
reV-NRWAL analysis module.

This module runs reV data through the NRWAL compute library. This code was
first developed to use a custom offshore wind LCOE equation library but has
since been refactored to analyze any equation library in NRWAL.

Everything in this module operates on the spatiotemporal resolution of the reV
generation output file. This is usually the wind or solar resource resolution
but could be the supply curve resolution after representative profiles is run.
"""
import numpy as np
import pandas as pd
import logging
from warnings import warn

from reV.generation.generation import Gen
from reV.handlers.outputs import Outputs
from reV.utilities.exceptions import (DataShapeError,
                                      OffshoreWindInputWarning,
                                      OffshoreWindInputError)
from reV.utilities import log_versions


logger = logging.getLogger(__name__)


class RevNrwal:
    """Framework to handle reV-NRWAL analysis."""

    # Default columns from the site_data table to join to the output meta data
    DEFAULT_META_COLS = ('config', )

    def __init__(self, gen_fpath, site_data, sam_files, nrwal_configs,
                 output_request, save_raw=True, meta_gid_col='gid',
                 site_meta_cols=None,
                 ):
        """
        Parameters
        ----------
        gen_fpath : str
            Full filepath to reV generation or rep_profiles h5 output file.
            Anything in the output_request is added and/or manipulated in this
            file.
        site_data : str | pd.DataFrame
            Site-specific input data for NRWAL calculation. String should be a
            filepath that points to a csv, DataFrame is pre-extracted data.
            Rows match sites, columns are input keys. Need a "gid" column that
            corresponds to the "meta_gid_col" in the gen_fpath meta data and a
            "config" column that corresponds to the nrwal_configs input. Only
            sites with a gid in this file's "gid" column will be run through
            NRWAL.
        sam_files : dict
            Dictionary lookup of config_id (keys) mapped to config filepaths
            (values). The same config_id values will be used from the
            nrwal_configs lookup input.
        nrwal_configs : dict
            Dictionary lookup of config_id (keys) mapped to config filepaths
            (values). The same config_id values will be used from the
            sam_files lookup in project_points
        output_request : list | tuple
            List of output dataset names you want written to the gen_fpath
            file. Any key from the NRWAL configs or any of the inputs
            (site_data or sam_files) is available to be exported as an output
            dataset. If you want to manipulate a dset like cf_mean from
            gen_fpath and include it in the output_request, you should set
            save_raw=True and then in the NRWAL equations use cf_mean_raw as
            the input and then define cf_mean as the manipulated data that will
            be included in the output_request.
        save_raw : bool
            Flag to save a copy of existing datasets in gen_fpath that are part
            of the output_request. For example, if you request cf_mean in
            output_request and manipulate the cf_mean dataset in the NRWAL
            equations, the original cf_mean will be archived under the
            "cf_mean_raw" dataset in gen_fpath.
        meta_gid_col : str
            Column label in the source meta data from gen_fpath that contains
            the unique gid identifier. This will be joined to the site_data
            "gid" column
        site_meta_cols : list | tuple | None
            Column labels from site_data to be added to the meta data table in
            gen_fpath. None (default) will use class variable
            DEFAULT_META_COLS, and any additional cols requested here will be
            added to DEFAULT_META_COLS.
        """

        log_versions(logger)

        # delayed NRWAL import to cause less errors with old reV installs
        # if not running nrwal.
        from NRWAL import NrwalConfig

        self._meta_gid_col = meta_gid_col
        self._gen_fpath = gen_fpath
        self._site_data = site_data
        self._output_request = output_request
        self._meta_out = None
        self._time_index = None
        self._save_raw = save_raw

        self._nrwal_configs = {k: NrwalConfig(v) for k, v in
                               nrwal_configs.items()}

        self._site_meta_cols = site_meta_cols
        if self._site_meta_cols is None:
            self._site_meta_cols = list(self.DEFAULT_META_COLS)
        else:
            self._site_meta_cols = list(self._site_meta_cols)
            self._site_meta_cols += list(self.DEFAULT_META_COLS)
            self._site_meta_cols = list(set(self._site_meta_cols))

        self._site_data = self._parse_site_data()
        self._meta_source = self._parse_gen_data()
        self._analysis_gids, self._site_data = self._parse_analysis_gids()

        pc = Gen.get_pc(self._site_data[['gid', 'config']], points_range=None,
                        sam_configs=sam_files, tech='windpower')
        self._project_points = pc.project_points

        self._sam_sys_inputs = self._parse_sam_sys_inputs()
        self._preflight_checks()
        self.save_raw_dsets()
        self._nrwal_inputs = self._get_input_data()
        self._out = self._init_outputs()

        meta_gids = self.meta_source[self._meta_gid_col].values
        logger.info('Finished initializing NRWAL analysis module for "{}" '
                    '{} through {} with {} total generation points and '
                    '{} NRWAL analysis points.'
                    .format(self._meta_gid_col, meta_gids.min(),
                            meta_gids.max(), len(self.meta_source),
                            len(self.analysis_gids)))

    def _parse_site_data(self, required_columns=('gid', 'config')):
        """Parse the site-specific spatial input data file

        Parameters
        ----------
        required_columns : tuple | list
            List of column names that must be in the site_data in
            order to run the reV NRWAL module.

        Returns
        -------
        site_data : pd.DataFrame
            Dataframe of extracted site_data. Each row is an analysis point and
            columns are spatial data inputs.
        """

        if isinstance(self._site_data, str):
            self._site_data = pd.read_csv(self._site_data)

        if 'dist_l_to_ts' in self._site_data:
            if self._site_data['dist_l_to_ts'].sum() > 0:
                w = ('Possible incorrect Offshore data input! "dist_l_to_ts" '
                     '(distance land to transmission) input is non-zero. '
                     'Most reV runs set this to zero and input the cost '
                     'of transmission from landfall tie-in to '
                     'transmission feature in the supply curve module.')
                logger.warning(w)
                warn(w, OffshoreWindInputWarning)

        for c in required_columns:
            if c not in self._site_data:
                msg = ('Did not find required "{}" column in site_data!'
                       .format(c))
                logger.error(msg)
                raise KeyError(msg)

        self._site_data = self._site_data.sort_values('gid')

        return self._site_data

    def _parse_gen_data(self):
        """Parse generation data and get meta data

        Returns
        -------
        meta : pd.DataFrame
            Full meta data from gen_fpath.
        """

        with Outputs(self._gen_fpath, mode='r') as out:
            meta = out.meta

        msg = ('Could not find "{}" column in source generation h5 file '
               'meta data! Available cols: {}'
               .format(self._meta_gid_col, meta.columns.values.tolist()))
        assert self._meta_gid_col in meta, msg

        # currently an assumption of sorted gids in the reV gen output
        msg = ('Source capacity factor meta data is not ordered!')
        meta_gids = list(meta[self._meta_gid_col])
        assert meta_gids == sorted(meta_gids), msg

        return meta

    def _parse_analysis_gids(self):
        """Check the intersection of the generation gids and the site_data
        input gids.

        Returns
        -------
        analysis_gids : np.ndarray
            Array indicating which sites in the source meta data to process
            with NRWAL. This is the intersection of the gids in the generation
            meta data and the gids in the site_data input.
        site_data : pd.DataFrame
            The site_data table reduced to only those gids that are in the
            analysis_gids
        """

        meta_gids = self.meta_source[self._meta_gid_col].values

        missing = ~np.isin(meta_gids, self._site_data['gid'])
        if any(missing):
            msg = ('{} sites from the generation meta data input were '
                   'missing from the "site_data" input and will not be '
                   'run through NRWAL: {}'
                   .format(missing.sum(), meta_gids[missing]))
            logger.info(msg)

        missing = ~np.isin(self._site_data['gid'], meta_gids)
        if any(missing):
            missing = self._site_data['gid'].values[missing]
            msg = ('{} sites from the "site_data" input were missing from the '
                   'generation meta data and will not be run through NRWAL: {}'
                   .format(len(missing), missing))
            logger.info(msg)

        analysis_gids = set(meta_gids) & set(self._site_data['gid'])
        analysis_gids = np.array(sorted(list(analysis_gids)))

        # reduce the site data table to only those sites being analyzed
        mask = np.isin(self._site_data['gid'], meta_gids)
        self._site_data = self._site_data[mask]

        return analysis_gids, self._site_data

    def _parse_sam_sys_inputs(self):
        """Get the SAM system inputs dict from project points.

        Returns
        -------
        system_inputs : pd.DataFrame
            DataFrame of SAM config inputs (columns) for every active nrwal
            analysis gid (row). Index is resource gids and there is also a
            column "gid" with the copied gids.
        """

        system_inputs = {}

        for gid in self.analysis_gids:
            system_inputs[gid] = self._project_points[gid][1]

        system_inputs = pd.DataFrame(system_inputs).T
        system_inputs = system_inputs.sort_index()
        system_inputs['gid'] = system_inputs.index.values
        system_inputs.index.name = 'gid'
        mask = system_inputs['gid'].isin(self.analysis_gids)
        system_inputs = system_inputs[mask]

        return system_inputs

    def _init_outputs(self):
        """Initialize a dictionary of outputs with dataset names as keys and
        numpy arrays as values. All datasets are initialized as 1D arrays and
        must be overwritten if found to be 2D. Only active analysis sites will
        have data in the output, sites that were not found in the site_data
        "gid" column will not have data in these output arrays

        Returns
        -------
        out : dict
            Dictionary of output data
        """
        out = {}

        for key in self._output_request:
            out[key] = np.full(len(self.analysis_gids), np.nan,
                               dtype=np.float32)

            if key in self.gen_dsets and not self._save_raw:
                msg = ('Output request "{0}" was also found in '
                       'the source gen file but save_raw=False! If '
                       'you are manipulating this '
                       'dset, make sure you set save_raw=False '
                       'and reference "{0}_raw" as the '
                       'input in the NRWAL equations and then define "{0}" '
                       'as the final manipulated dataset.'.format(key))
                logger.warning(msg)
                warn(msg)
            elif key in self.gen_dsets:
                msg = ('Output request "{0}" was also found in '
                       'the source gen file. If you are manipulating this '
                       'dset, make sure you reference "{0}_raw" as the '
                       'input in the NRWAL equations and then define "{0}" '
                       'as the final manipulated dataset.'.format(key))
                logger.info(msg)

            if key in self._nrwal_inputs:
                out[key] = self._nrwal_inputs[key]

        return out

    def _preflight_checks(self):
        """Run some preflight checks on the offshore inputs"""
        sam_files = {k: v for k, v in
                     self._project_points.sam_inputs.items()
                     if k in self._nrwal_configs}

        for cid, sys_in in sam_files.items():
            loss1 = sys_in.get('wind_farm_losses_percent', 0)
            loss2 = sys_in.get('turb_generic_loss', 0)
            if loss1 != 0 or loss2 != 0:
                msg = ('Wind farm loss for config "{}" is not 0. When using '
                       'NRWAL for offshore analysis, consider using gross '
                       'capacity factors from reV generation and applying '
                       'spatially dependent losses from the NRWAL equations'
                       .format(cid))
                logger.info(msg)

        available_ids = list(self._nrwal_configs.keys())
        requested_ids = list(self._site_data['config'].values)
        missing = set(requested_ids) - set(available_ids)
        if any(missing):
            msg = ('The following config ids were requested in the offshore '
                   'data input but were not available in the NRWAL config '
                   'input dict: {}'.format(missing))
            logger.error(msg)
            raise OffshoreWindInputError(msg)

        check_gid_order = (self._site_data['gid'].values
                           == self._sam_sys_inputs['gid'].values)
        msg = 'NRWAL site_data and system input dataframe had bad order'
        assert (check_gid_order).all(), msg

        missing = [c for c in self._site_meta_cols if c not in self._site_data]
        if any(missing):
            msg = ('Could not find requested NRWAL site data pass through '
                   'columns in offshore input data: {}'.format(missing))
            logger.error(msg)
            raise OffshoreWindInputError(msg)

    def _get_input_data(self):
        """Get all the input data from the site_data, SAM system configs, and
        generation h5 file, formatted together in one dictionary for NRWAL.

        Returns
        -------
        nrwal_inputs : dict
            Dictionary mapping required NRWAL input variable names (keys) to 1
            or 2D arrays of inputs for all the analysis_gids
        """

        logger.info('Setting up input data for NRWAL...')

        # preconditions for this to work properly
        assert len(self._site_data) == len(self.analysis_gids)
        assert len(self._sam_sys_inputs) == len(self.analysis_gids)

        all_required = []
        for config_id, nrwal_config in self._nrwal_configs.items():
            all_required += list(nrwal_config.required_inputs)
            all_required = list(set(all_required))

            missing_vars = [var for var in nrwal_config.required_inputs
                            if var not in self._site_data
                            and var not in self.meta_source
                            and var not in self._sam_sys_inputs
                            and var not in self.gen_dsets]

            if any(missing_vars):
                msg = ('Could not find required input variables {} '
                       'for NRWAL config "{}" in either the offshore '
                       'data or the SAM system data!'
                       .format(missing_vars, config_id))
                logger.error(msg)
                raise OffshoreWindInputError(msg)

        meta_data_vars = [var for var in all_required
                          if var in self.meta_source]
        logger.info('Pulling the following inputs from the gen meta data: {}'
                    .format(meta_data_vars))
        nrwal_inputs = {var: self.meta_source[var].values[self.analysis_mask]
                        for var in meta_data_vars}

        site_data_vars = [var for var in all_required
                          if var in self._site_data
                          and var not in nrwal_inputs]
        site_data_vars.append('config')
        logger.info('Pulling the following inputs from the site_data input: {}'
                    .format(site_data_vars))
        for var in site_data_vars:
            nrwal_inputs[var] = self._site_data[var].values

        sam_sys_vars = [var for var in all_required
                        if var in self._sam_sys_inputs
                        and var not in nrwal_inputs]
        logger.info('Pulling the following inputs from the SAM system '
                    'configs: {}'.format(sam_sys_vars))
        for var in sam_sys_vars:
            nrwal_inputs[var] = self._sam_sys_inputs[var].values

        gen_vars = [var for var in all_required
                    if var in self.gen_dsets
                    and var not in nrwal_inputs]
        logger.info('Pulling the following inputs from the generation '
                    'h5 file: {}'.format(gen_vars))
        with Outputs(self._gen_fpath, mode='r') as f:
            source_gids = self.meta_source[self._meta_gid_col]
            gen_gids = np.where(source_gids.isin(self.analysis_gids))[0]
            for var in gen_vars:
                shape = f.shapes[var]
                if len(shape) == 1:
                    nrwal_inputs[var] = f[var, gen_gids]
                elif len(shape) == 2:
                    nrwal_inputs[var] = f[var, :, gen_gids]
                else:
                    msg = ('Data shape for "{}" must be 1 or 2D but '
                           'received: {}'.format(var, shape))
                    logger.error(msg)
                    raise DataShapeError(msg)

        logger.info('Finished setting up input data for NRWAL!')

        return nrwal_inputs

    @property
    def time_index(self):
        """Get the source time index."""
        if self._time_index is None:
            with Outputs(self._gen_fpath, mode='r') as out:
                self._time_index = out.time_index

        return self._time_index

    @property
    def gen_dsets(self):
        """Get the available datasets from the gen source file"""
        with Outputs(self._gen_fpath, mode='r') as out:
            dsets = out.dsets

        return dsets

    @property
    def meta_source(self):
        """Get the full meta data (onshore + offshore)"""
        return self._meta_source

    @property
    def meta_out(self):
        """Get the combined onshore and offshore meta data."""
        if self._meta_out is None:
            self._meta_out = self._meta_source.copy()
            for col in self._site_meta_cols:
                data = self._nrwal_inputs[col]
                self._meta_out.loc[self.analysis_mask, col] = data

        return self._meta_out

    @property
    def analysis_mask(self):
        """Get a boolean array to mask the source generation meta data where
        True is sites that are to be analyzed by NRWAL.

        Returns
        -------
        np.ndarray
        """
        mask = np.isin(self.meta_source[self._meta_gid_col],
                       self.analysis_gids)
        return mask

    @property
    def analysis_gids(self):
        """Get an array of gids from the source generation meta data that are
        to-be analyzed by nrwal.

        Returns
        -------
        np.ndarray
        """
        return self._analysis_gids

    @property
    def outputs(self):
        """Get a dict of NRWAL outputs. Only active analysis sites will have
        data in the output, sites that were not found in the site_data "gid"
        column will not have data in these output arrays"""
        return self._out

    def _save_nrwal_out(self, name, nrwal_out, output_mask):
        """Save a dataset from the nrwal_out dictionary to the self._out
        attribute

        Parameters
        ----------
        name : str
            Dataset name of the nrwal output to be saved.
        nrwal_out : dict
            Output dictionary from a successfully evaluated NrwalConfig object
            containing the dataset with the input name
        output_mask : np.ndarray
            Boolean array showing which gids in self.analysis_gids should be
            assigned data from this NRWAL output. If not all true, there are
            probably multiple NrwalConfig objects that map to different sets of
            gids.
        """
        value = nrwal_out[name]
        value = self._value_to_array(value, name)

        if len(value.shape) == 1:
            self._out[name][output_mask] = value[output_mask]

        elif len(value.shape) == 2:
            if len(self._out[name].shape) == 1:
                if not all(np.isnan(self._out[name])):
                    msg = ('Output dataset "{}" was initialized as 1D but was '
                           'later found to be 2D but was not all NaN!'
                           .format(name))
                    logger.error(msg)
                    raise DataShapeError(msg)

                # re-initialize the dataset as 2D now that we
                # know what the output looks like
                out_shape = (len(self.time_index), len(self.analysis_gids))
                self._out[name] = np.full(out_shape, np.nan, dtype=np.float32)

            self._out[name][:, output_mask] = value[:, output_mask]

        else:
            msg = ('Could not make sense of NRWAL output "{}" '
                   'with shape {}'.format(name, value.shape))
            logger.error(msg)
            raise DataShapeError(msg)

    def _save_nrwal_misc(self, name, nrwal_config, output_mask):
        """Save miscellaneous output requests from a NRWAL config object (not
        NRWAL output dictionary) to the self._out attribute.

        Parameters
        ----------
        name : str
            Dataset name of the nrwal output to be saved.
        nrwal_config : NrwalConfig
            NrwalConfig object containing NRWAL Equation objects or something
            else that is to be exported to the outputs
        output_mask : np.ndarray
            Boolean array showing which gids in self.analysis_gids should be
            assigned data from this NRWAL output. If not all true, there are
            probably multiple NrwalConfig objects that map to different sets of
            gids.
        """

        from NRWAL import Equation
        value = nrwal_config[name]

        if isinstance(value, Equation):
            msg = ('Cannot retrieve Equation "{}" from NRWAL. '
                   'Must be a number!'.format(name))
            assert not any(value.variables), msg
            value = value.eval()

        value = self._value_to_array(value, name)
        self._out[name][output_mask] = value[output_mask]

    def _value_to_array(self, value, name):
        """Turn the input into numpy array if it isn't already."""
        if np.issubdtype(type(value), np.number):
            value *= np.ones(len(self.analysis_gids), dtype=np.float32)

        if not isinstance(value, np.ndarray):
            msg = ('NRWAL key "{}" returned bad type of "{}", needs to be '
                   'numeric or an output array.'.format(name, type(value)))
            logger.error(msg)
            raise TypeError(msg)
        return value

    def run_nrwal(self):
        """Run analysis via the NRWAL analysis library"""

        for i, (cid, nrwal_config) in enumerate(self._nrwal_configs.items()):
            output_mask = self._site_data['config'].values == cid
            logger.info('Running NRWAL config {} of {}: "{}" and applying '
                        'to {} out of {} total sites'
                        .format(i + 1, len(self._nrwal_configs), cid,
                                output_mask.sum(), len(output_mask)))

            nrwal_out = nrwal_config.eval(inputs=self._nrwal_inputs)

            # pylint: disable=C0201
            for name in self._out.keys():
                if name in nrwal_out:
                    self._save_nrwal_out(name, nrwal_out, output_mask)

                elif name in nrwal_config.keys():
                    self._save_nrwal_misc(name, nrwal_config, output_mask)

                elif name not in self._nrwal_inputs:
                    msg = ('Could not find "{}" in the output dict of NRWAL '
                           'config {}'.format(name, cid))
                    logger.warning(msg)
                    warn(msg)

    def check_outputs(self):
        """Check the nrwal outputs for nan values and raise errors if found."""
        for name, arr in self._out.items():
            if np.isnan(arr).all():
                msg = ('Output array "{}" is all NaN! Probably was not '
                       'found in the available NRWAL keys.'.format(name))
                logger.warning(msg)
                warn(msg)
            elif np.isnan(arr).any():
                mask = np.isnan(arr)
                nan_meta = self.meta_source[self.analysis_mask][mask]
                nan_gids = nan_meta[self._meta_gid_col].values
                msg = ('NaN values ({} out of {}) persist in NRWAL '
                       'output "{}"!'
                       .format(np.isnan(arr).sum(), len(arr), name))
                logger.warning(msg)
                logger.warning('This is the NRWAL meta that is causing NaN '
                               'outputs: {}'.format(nan_meta))
                logger.warning('These are the resource gids causing NaN '
                               'outputs: {}'.format(nan_gids))
                warn(msg)

    def save_raw_dsets(self):
        """If requested by save_raw=True, archive raw datasets that exist in
        the gen_fpath file and are also requested in the output_request"""
        if self._save_raw:
            with Outputs(self._gen_fpath, 'a') as f:
                for dset in self._output_request:
                    dset_raw = '{}_raw'.format(dset)
                    if dset in f and dset_raw not in f:
                        logger.info('Saving raw data from "{}" to "{}"'
                                    .format(dset, dset_raw))
                        f._add_dset(dset_raw, f[dset], f.dtypes[dset],
                                    attrs=f.attrs[dset])

    def write_to_gen_fpath(self):
        """Save NRWAL outputs to input generation fpath file."""

        logger.info('Writing NRWAL outputs to: {}'.format(self._gen_fpath))
        write_all = self.analysis_mask.all()

        with Outputs(self._gen_fpath, 'a') as f:
            meta_attrs = f.attrs['meta']
            del f._h5['meta']
            f._set_meta('meta', self.meta_out, attrs=meta_attrs)

            for dset, arr in self._out.items():
                if len(arr.shape) == 1:
                    data = np.full(len(self.meta_source), np.nan,
                                   dtype=np.float32)
                else:
                    full_shape = (len(self.time_index),
                                  len(self.meta_source))
                    data = np.full(full_shape, np.nan, dtype=np.float32)

                dset_attrs = {'scale_factor': 1}
                dset_dtype = np.float32
                if dset in f.dsets:
                    logger.info('Found "{}" in file, loading data and '
                                'overwriting data for {} out of {} sites.'
                                .format(dset, self.analysis_mask.sum(),
                                        len(self.analysis_mask)))
                    dset_attrs = f.attrs[dset]
                    dset_dtype = f.dtypes[dset]
                    if not write_all:
                        data = f[dset]

                if len(arr.shape) == 1:
                    data[self.analysis_mask] = arr
                else:
                    data[:, self.analysis_mask] = arr

                logger.info('Writing final "{}" to: {}'
                            .format(dset, self._gen_fpath))
                f._add_dset(dset, data, dset_dtype, attrs=dset_attrs)

        logger.info('Finished writing NRWAL outputs to: {}'
                    .format(self._gen_fpath))

    @classmethod
    def run(cls, gen_fpath, site_data, sam_files, nrwal_configs,
            output_request, save_raw=True, meta_gid_col='gid',
            site_meta_cols=None):
        """Initialize and run the NRWAL analysis object.

        Parameters
        ----------
        gen_fpath : str
            Full filepath to reV generation or rep_profiles h5 output file.
            Anything in the output_request is added and/or manipulated in this
            file.
        site_data : str | pd.DataFrame
            Site-specific input data for NRWAL calculation. String should be a
            filepath that points to a csv, DataFrame is pre-extracted data.
            Rows match sites, columns are input keys. Need a "gid" column that
            corresponds to the "meta_gid_col" in the gen_fpath meta data and a
            "config" column that corresponds to the nrwal_configs input. Only
            sites with a gid in this file's "gid" column will be run through
            NRWAL.
        sam_files : dict
            Dictionary lookup of config_id (keys) mapped to config filepaths
            (values). The same config_id values will be used from the
            nrwal_configs lookup input.
        nrwal_configs : dict
            Dictionary lookup of config_id (keys) mapped to config filepaths
            (values). The same config_id values will be used from the
            sam_files lookup in project_points
        output_request : list | tuple
            List of output dataset names you want written to the gen_fpath
            file. Any key from the NRWAL configs or any of the inputs
            (site_data or sam_files) is available to be exported as an output
            dataset. If you want to manipulate a dset like cf_mean from
            gen_fpath and include it in the output_request, you should set
            save_raw=True and then in the NRWAL equations use cf_mean_raw as
            the input and then define cf_mean as the manipulated data that will
            be included in the output_request.
        save_raw : bool
            Flag to save a copy of existing datasets in gen_fpath that are part
            of the output_request. For example, if you request cf_mean in
            output_request and manipulate the cf_mean dataset in the NRWAL
            equations, the original cf_mean will be archived under the
            "cf_mean_raw" dataset in gen_fpath.
        meta_gid_col : str
            Column label in the source meta data from gen_fpath that contains
            the unique gid identifier. This will be joined to the site_data
            "gid" column
        site_meta_cols : list | tuple | None
            Column labels from site_data to be added to the meta data table in
            gen_fpath. None (default) will use class variable
            DEFAULT_META_COLS, and any additional cols requested here will be
            added to DEFAULT_META_COLS.

        Returns
        -------
        obj : RevNrwal
            Instantiated and run RevNrwal analysis object.
        """

        obj = cls(gen_fpath, site_data, sam_files, nrwal_configs,
                  output_request,
                  save_raw=save_raw,
                  meta_gid_col=meta_gid_col,
                  site_meta_cols=site_meta_cols)

        if any(obj.analysis_gids):
            obj.run_nrwal()
            obj.check_outputs()
            obj.write_to_gen_fpath()

        logger.info('NRWAL module complete!')

        return obj



# -*- coding: utf-8 -*-
"""Collection of functions used to hybridize columns in rep profiles meta.

@author: ppinchuk
"""


def aggregate_solar_capacity(h):
    """Compute the total solar capcity allowed in hybridization.

    Note
    ----
    No limiting is done on the ratio of wind to solar. This method
    checks for an existing 'hybrid_solar_capacity'. If one does not exist,
    it is assumed that there is no limit on the solar to wind capacity
    ratio and the solar capacity is copied into this new column.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the capacity allowed in the hybrid
        capacity sum, or `None` if 'hybrid_solar_capacity' already exists.

    Notes
    -----

    """
    if 'hybrid_solar_capacity' in h.hybrid_meta:
        return None
    return h.hybrid_meta['solar_capacity']


def aggregate_wind_capacity(h):
    """Compute the total wind capcity allowed in hybridization.

    Note
    ----
    No limiting is done on the ratio of wind to solar. This method
    checks for an existing 'hybrid_wind_capacity'. If one does not exist,
    it is assumed that there is no limit on the solar to wind capacity
    ratio and the wind capacity is copied into this new column.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the capacity allowed in the hybrid
        capacity sum, or `None` if 'hybrid_solar_capacity' already exists.

    Notes
    -----

    """
    if 'hybrid_wind_capacity' in h.hybrid_meta:
        return None
    return h.hybrid_meta['wind_capacity']


def aggregate_capacity(h):
    """Compute the total capcity by summing the individual capacities.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the aggregated capacity, or `None`
        if the capacity columns are missing.
    """

    sc, wc = 'hybrid_solar_capacity', 'hybrid_wind_capacity'
    missing_solar_cap = sc not in h.hybrid_meta.columns
    missing_wind_cap = wc not in h.hybrid_meta.columns
    if missing_solar_cap or missing_wind_cap:
        return None

    total_cap = h.hybrid_meta[sc] + h.hybrid_meta[wc]
    return total_cap


def aggregate_capacity_factor(h):
    """Compute the capacity-weighted mean capcity factor.

    Parameters
    ----------
    h : `reV.hybrids.Hybridization`
        Instance of `reV.hybrids.Hybridization` class containing the
        attribute `hybrid_meta`, which is a DataFrame containing
        hybridized meta data.

    Returns
    -------
    data : Series | None
        A series of data containing the aggregated capacity, or `None`
        if the capacity and/or mean_cf columns are missing.
    """

    sc, wc = 'hybrid_solar_capacity', 'hybrid_wind_capacity'
    scf, wcf = 'solar_mean_cf', 'wind_mean_cf'
    missing_solar_cap = sc not in h.hybrid_meta.columns
    missing_wind_cap = wc not in h.hybrid_meta.columns
    missing_solar_mean_cf = scf not in h.hybrid_meta.columns
    missing_wind_mean_cf = wcf not in h.hybrid_meta.columns
    missing_any = (missing_solar_cap or missing_wind_cap
                   or missing_solar_mean_cf or missing_wind_mean_cf)
    if missing_any:
        return None

    solar_cf_weighted = h.hybrid_meta[sc] * h.hybrid_meta[scf]
    wind_cf_weighted = h.hybrid_meta[wc] * h.hybrid_meta[wcf]
    total_capacity = aggregate_capacity(h)
    hybrid_cf = (solar_cf_weighted + wind_cf_weighted) / total_capacity
    return hybrid_cf


HYBRID_METHODS = {
    'hybrid_solar_capacity': aggregate_solar_capacity,
    'hybrid_wind_capacity': aggregate_wind_capacity,
    'hybrid_capacity': aggregate_capacity,
    'hybrid_mean_cf': aggregate_capacity_factor
}



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Generation CLI entry points.
"""
import click
import logging
from math import ceil
import os
import pprint
import time
from warnings import warn

from reV.config.cli_project_points import _parse_lat_lons, _parse_regions
from reV.config.project_points import ProjectPoints, PointsControl
from reV.config.sam_analysis_configs import GenConfig
from reV.generation.generation import Gen
from reV.pipeline.status import Status
from reV.utilities.exceptions import ConfigError, ProjectPointsValueError
from reV.utilities.cli_dtypes import SAMFILES, PROJECTPOINTS
from reV.utilities import ModuleName
from reV import __version__

from rex.utilities.cli_dtypes import INT, STR, INTLIST, STRLIST
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_mult
from rex.utilities.utilities import parse_year, get_class_properties

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=os.path.basename(os.getcwd()),
              type=STR, show_default=True, help='reV Economics job name.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Generation Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid Gen config keys
    """
    click.echo(', '.join(get_class_properties(GenConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV generation configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV gen from a config file."""
    verbose = any([verbose, ctx.obj['VERBOSE']])

    # Instantiate the config object
    config = GenConfig(config_file)

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # make output directory if does not exist
    if not os.path.exists(config.dirout):
        os.makedirs(config.dirout)

    # initialize loggers.
    init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV Generation from config file: "{}"'
                .format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))
    logger.info('The following project points were specified: "{}"'
                .format(config.get('project_points', None)))
    logger.info('The following SAM configs are available to this run:\n{}'
                .format(pprint.pformat(config.get('sam_files', None),
                                       indent=4)))
    logger.info('The following is being used for site specific input data: '
                '"{}"'.format(config.site_data))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    # set config objects to be passed through invoke to direct methods
    ctx.obj['TECH'] = config.technology
    ctx.obj['POINTS'] = config.project_points
    ctx.obj['SAM_FILES'] = config.sam_files
    ctx.obj['LOGDIR'] = config.log_directory
    ctx.obj['OUTPUT_REQUEST'] = config.output_request
    ctx.obj['GID_MAP'] = config.gid_map
    ctx.obj['SITE_DATA'] = config.site_data
    ctx.obj['TIMEOUT'] = config.timeout
    ctx.obj['SITES_PER_WORKER'] = config.execution_control.sites_per_worker
    ctx.obj['MAX_WORKERS'] = config.execution_control.max_workers
    ctx.obj['MEM_UTIL_LIM'] = \
        config.execution_control.memory_utilization_limit

    ctx.obj['CURTAILMENT'] = None
    if config.curtailment is not None:
        # pass through the curtailment file, not the curtailment object
        ctx.obj['CURTAILMENT'] = config['curtailment']

    for i, year in enumerate(config.analysis_years):
        submit_from_config(ctx, name, year, config, i, verbose=verbose)


def submit_from_config(ctx, name, year, config, i, verbose=False):
    """Function to submit one year from a config file.

    Parameters
    ----------
    ctx : cli.ctx
        Click context object. Use case: data = ctx.obj['key']
    name : str
        Job name.
    year : int | str | NoneType
        4 digit year or None.
    config : reV.config.GenConfig
        Generation config object.
    i : int
        Index variable associated with the index of the year in analysis years.
    verbose : bool
        Flag to turn on debug logging. Default is not verbose.
    """
    res_files = config.parse_res_files()
    # set the year-specific variables
    ctx.obj['RES_FILE'] = res_files[i]

    # check to make sure that the year matches the resource file
    if str(year) not in res_files[i]:
        warn('Resource file and year do not appear to match. '
             'Expected the string representation of the year '
             'to be in the resource file name. '
             'Year: {}, Resource file: {}'
             .format(year, res_files[i]))

    # if the year isn't in the name, add it before setting the file output
    fout = make_fout(name, year)
    ctx.obj['OUT_FPATH'] = os.path.join(config.dirout, fout)

    # invoke direct methods based on the config execution option
    if config.execution_control.option == 'local':
        name_year = make_fout(name, year).replace('.h5', '')
        ctx.obj['NAME'] = name_year
        status = Status.retrieve_job_status(
            config.dirout, module=ModuleName.GENERATION, job_name=name_year)
        if status == 'successful':
            logger.info('reV Generation job with name "{}" was already '
                        'successfully run in directory: {}'
                        .format(name, config.out_dir))
        else:
            job_attrs = {'hardware': 'local', 'fout': fout,
                         'dirout': config.dirout}
            Status.add_job(
                config.dirout, module=ModuleName.GENERATION,
                job_name=name_year, replace=True,
                job_attrs=job_attrs)
            ctx.invoke(local,
                       max_workers=config.execution_control.max_workers,
                       timeout=config.timeout,
                       points_range=None,
                       verbose=verbose)

    elif config.execution_control.option in ('eagle', 'slurm'):
        if not parse_year(name, option='bool') and year:
            # Add year to name before submitting
            ctx.obj['NAME'] = '{}_{}'.format(name, str(year))
        ctx.invoke(slurm, nodes=config.execution_control.nodes,
                   alloc=config.execution_control.allocation,
                   walltime=config.execution_control.walltime,
                   memory=config.execution_control.memory,
                   feature=config.execution_control.feature,
                   conda_env=config.execution_control.conda_env,
                   module=config.execution_control.module,
                   stdout_path=os.path.join(config.log_directory, 'stdout'),
                   sh_script=config.execution_control.sh_script,
                   verbose=verbose)


def make_fout(name, year):
    """Make an appropriate file output from name and year.

    Parameters
    ----------
    name : str
        Job name.
    year : int | str
        Analysis year.

    Returns
    -------
    fout : str
        .h5 output file based on name and year
    """

    try:
        match = parse_year(name)
    except RuntimeError:
        match = False

    # if the year isn't in the name, add it before setting the file output
    if match and year:
        if str(year) != str(match):
            raise ConfigError('Tried to submit gen job for {}, but found a '
                              'different year in the base job name: "{}". '
                              'Please remove the year from the job name.'
                              .format(year, name))
    if year:
        fout = '{}{}.h5'.format(name, '_{}'.format(year) if not
                                match else '')
    else:
        fout = '{}.h5'.format(name)
    return fout


@main.group()
@click.option('--tech', '-t', required=True, type=STR,
              help='reV tech to analyze (required).')
@click.option('--sam_files', '-sf', required=True, type=SAMFILES,
              help='SAM config files (required) (str, dict, or list).')
@click.option('--res_file', '-rf', required=True,
              help='Filepath to single resource file, multi-h5 directory, '
              'or /h5_dir/prefix*suffix.')
@click.option('--out_fpath', '-o', type=click.Path(), required=True,
              help='Ouput .h5 file path')
@click.option('--points', '-p', default=None, type=PROJECTPOINTS,
              show_default=True,
              help=('reV project points to analyze '
                    '(slice, list, or file string). '
                    'Default is slice(0, 100)'))
@click.option('--lat_lon_fpath', '-llf', type=click.Path(exists=True),
              default=None, show_default=True,
              help=('File path to .csv or .json containing latitude, '
                    'longitude coordinates of interest'))
@click.option('--lat_lon_coords', '--llc', nargs=2, type=float, default=None,
              show_default=True,
              help='(lat, lon) coordinates of interest')
@click.option('--regions', '-regs', type=STR, default=None,
              show_default=True,
              help='File path to .json containing regions of interest')
@click.option('--region', '-r', type=STR, default=None,
              show_default=True,
              help='Region to extract')
@click.option('--region_col', '-col', type=STR, default='state',
              show_default=True,
              help='Meta column to search for region')
@click.option('--sites_per_worker', '-spw', default=None, type=INT,
              show_default=True,
              help=('Number of sites to run in series on a single worker. '
                    'Default is the resource column chunk size.'))
@click.option('--logdir', '-lo', default='./out/log_gen', type=STR,
              help='Generation log file directory. Default is ./out/log_gen')
@click.option('--output_request', '-or', type=STRLIST, default=['cf_mean'],
              show_default=True,
              help=('List of requested output variable names. '
                    'Default is ["cf_mean"].'))
@click.option('--site_data', '-sd', default=None, type=click.Path(exists=True),
              show_default=True,
              help='Site-specific data file for gen calculation. Input '
              'should be a filepath that points to a csv. Rows match sites, '
              'columns are input keys. Needs a "gid" column. Input as None '
              'if no site-specific data. Can be the same csv as the '
              'project_points input')
@click.option('--mem_util_lim', '-mem', type=float, default=0.4,
              show_default=True,
              help='Fractional node memory utilization limit. Default is 0.4 '
              'to account for numpy memory spikes and memory bloat.')
@click.option('--curtailment', '-curt', type=click.Path(exists=True),
              default=None, show_default=True,
              help='Optional JSON file with curtailment inputs parameters.')
@click.option('--gid_map', '-gm', type=click.Path(exists=True),
              default=None, show_default=True,
              help='Optional gid mapping that can be used to map unique '
              'generation gids to non-unique gids in the resource file. '
              'Should be a filepath to a csv with columns gid and gid_map. '
              'Can be the same csv as project_points input.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, tech, sam_files, res_file, out_fpath, points, lat_lon_fpath,
           lat_lon_coords, regions, region, region_col, sites_per_worker,
           logdir, output_request, site_data, mem_util_lim,
           curtailment, gid_map, verbose):
    """Run reV gen directly w/o a config file."""
    ctx.obj['TECH'] = tech
    ctx.obj['POINTS'] = points
    ctx.obj['SAM_FILES'] = sam_files
    ctx.obj['RES_FILE'] = res_file
    ctx.obj['SITES_PER_WORKER'] = sites_per_worker
    ctx.obj['OUT_FPATH'] = out_fpath
    ctx.obj['LOGDIR'] = logdir
    ctx.obj['OUTPUT_REQUEST'] = output_request
    ctx.obj['SITE_DATA'] = site_data
    ctx.obj['MEM_UTIL_LIM'] = mem_util_lim
    ctx.obj['CURTAILMENT'] = curtailment
    ctx.obj['GID_MAP'] = gid_map

    ctx.obj['LAT_LON_FPATH'] = lat_lon_fpath
    ctx.obj['LAT_LON_COORDS'] = lat_lon_coords
    ctx.obj['REGIONS'] = regions
    ctx.obj['REGION'] = region
    ctx.obj['REGION_COL'] = region_col

    verbose = any([verbose, ctx.obj['VERBOSE']])


def _parse_points(ctx):
    """
    Parse project points from CLI inputs

    Parameters
    ----------
    ctx : dict
        CLI context object

    Returns
    -------
    points : slice | list | string | ProjectPoints
        ProjectPoints or points to initialize ProjectPoints
    """
    tech = ctx.obj['TECH']
    points = ctx.obj['POINTS']
    sam_files = ctx.obj['SAM_FILES']
    res_file = ctx.obj['RES_FILE']
    curtailment = ctx.obj['CURTAILMENT']
    lat_lon_fpath = ctx.obj.get('LAT_LON_FPATH', None)
    lat_lon_coords = ctx.obj.get('LAT_LON_COORDS', None)
    regions = ctx.obj.get('REGIONS', None)
    region = ctx.obj.get('REGION', None)
    region_col = ctx.obj.get('REGION_COL', 'state')

    i = 0
    if points is not None:
        i += 1

    if lat_lon_fpath is not None or lat_lon_coords:
        lat_lons = _parse_lat_lons(lat_lon_fpath, lat_lon_coords)
        points = ProjectPoints.lat_lon_coords(lat_lons, res_file,
                                              sam_files, tech=tech,
                                              curtailment=curtailment)
        i += 1

    if region is not None or regions is not None:
        regions = _parse_regions(regions, region, region_col)
        points = ProjectPoints.regions(regions, res_file, sam_files,
                                       tech=tech,
                                       curtailment=curtailment)
        i += 1

    msg = None
    if i == 0:
        msg = ("reV Gen requires one of 'points', 'lat-lon-fpath', "
               "'lat-lon-coords', 'regions', or 'region' and region-col' "
               "must be supplied to determine points to compute generation "
               "for!")
    elif i > 1:
        msg = ("reV Gen can only produce a unique set of Project Points for "
               "a single input value for ONE of 'points', 'lat-lon-fpath', "
               "'lat-lon-coords', 'regions', or 'region' and region-col'")

    if msg is not None:
        logger.error(msg)
        raise ProjectPointsValueError(msg)

    return points


@direct.command()
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help='Number of workers. Use 1 for serial, None for all cores.')
@click.option('--timeout', '-to', type=INT, default=1800,
              show_default=True,
              help='Number of seconds to wait for parallel generation run '
              'iterations to complete before returning zeros. '
              'Default is 1800 seconds.')
@click.option('--points_range', '-pr', default=None, type=INTLIST,
              show_default=True,
              help='Optional range list to run a subset of sites.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def local(ctx, max_workers, timeout, points_range, verbose):
    """Run generation on local worker(s)."""

    name = ctx.obj['NAME']
    tech = ctx.obj['TECH']
    points = ctx.obj['POINTS']
    sam_files = ctx.obj['SAM_FILES']
    res_file = ctx.obj['RES_FILE']
    sites_per_worker = ctx.obj['SITES_PER_WORKER']
    out_fpath = ctx.obj['OUT_FPATH']
    logdir = ctx.obj['LOGDIR']
    output_request = ctx.obj['OUTPUT_REQUEST']
    site_data = ctx.obj['SITE_DATA']
    mem_util_lim = ctx.obj['MEM_UTIL_LIM']
    curtailment = ctx.obj['CURTAILMENT']
    gid_map = ctx.obj['GID_MAP']
    verbose = any([verbose, ctx.obj['VERBOSE']])

    # initialize loggers for multiple modules
    init_mult(name, logdir, modules=[__name__, 'reV', 'rex'],
              verbose=verbose, node=True)

    for key, val in ctx.obj.items():
        logger.debug('ctx var passed to local method: "{}" : "{}" with type '
                     '"{}"'.format(key, val, type(val)))

    logger.info('Gen local is being run with with job name "{}" and resource '
                'file: {}. Target output path is: {}'
                .format(name, res_file, out_fpath))
    t0 = time.time()

    points = _parse_points(ctx)

    # Execute the Generation module with smart data flushing.
    Gen.reV_run(tech,
                points,
                sam_files,
                res_file,
                site_data=site_data,
                output_request=output_request,
                curtailment=curtailment,
                gid_map=gid_map,
                max_workers=max_workers,
                sites_per_worker=sites_per_worker,
                points_range=points_range,
                out_fpath=out_fpath,
                mem_util_lim=mem_util_lim,
                timeout=timeout)

    tmp_str = ' with points range {}'.format(points_range)
    dirout, fout = os.path.split(out_fpath)
    runtime = (time.time() - t0) / 60
    logger.info('Gen compute complete for project points "{0}"{1}. '
                'Time elapsed: {2:.2f} min. Target output dir: {3}'
                .format(points, tmp_str if points_range else '',
                        runtime, dirout))

    # add job to reV status file.
    status = {'dirout': dirout, 'fout': fout, 'job_status': 'successful',
              'runtime': runtime, 'finput': res_file}
    Status.make_job_file(dirout, ModuleName.GENERATION, name, status)


def get_node_pc(points, sam_files, tech, res_file, nodes):
    """Get a PointsControl object to be send to HPC nodes.

    Parameters
    ----------
    points : slice | str | list | tuple
        Slice/list specifying project points, string pointing to a project
        points csv.
    sam_files : dict | str | list
        SAM input configuration ID(s) and file path(s). Keys are the SAM
        config ID(s), top level value is the SAM path. Can also be a single
        config file str. If it's a list, it is mapped to the sorted list
        of unique configs requested by points csv.
    tech : str
        reV technology being executed.
    res_file : str
        Optional resource file to find maximum length of project points if
        points slice stop is None.
    nodes : int
        Number of nodes that the PointsControl object is being split to.

    Returns
    -------
    pc : reV.config.project_points.PointsControl
        PointsControl object to be iterated and send to HPC nodes.
    """

    if isinstance(points, (str, slice, list, tuple)):
        # create points control via points
        pp = ProjectPoints(points, sam_files, tech=tech, res_file=res_file)
        sites_per_node = ceil(len(pp) / nodes)
        pc = PointsControl(pp, sites_per_split=sites_per_node)
    else:
        raise TypeError('Generation Points input type is unrecognized: '
                        '"{}"'.format(type(points)))
    return pc


def get_node_name_fout(name, fout, i, pc, hpc='slurm'):
    """Make a node name and fout unique to the run name, year, and node number.

    Parameters
    ----------
    name : str
        Base node/job name
    fout : str
        Base file output name (no path) (with or without .h5 extension)
    i : int | None
        Node number. If None, only a single node is being used and no
        enumeration is necessary.
    pc : reV.config.PointsControl
        A PointsControl instance that i is enumerated from.
    hpc : str
        HPC job submission tool name (e.g. slurm or pbs). Affects job name.

    Returns
    -------
    node_name : str
        Base node name with _00 tag. 16 char max.
    fout_node : str
        Base file output name with _node00 tag.
    """

    if name.endswith('.h5'):
        name = name.replace('.h5', '')

    if not fout.endswith('.h5'):
        fout += '.h5'

    if i is None or len(pc) == 1:
        fout_node = fout
        node_name = fout

    else:
        if hpc == 'slurm':
            node_name = '{0}_{1:02d}'.format(name, i)
        elif hpc == 'pbs':
            # 13 chars for pbs, (lim is 16, -3 for "_ID")
            node_name = '{0}_{1:02d}'.format(name[:13], i)

        fout_node = fout.replace('.h5', '_node{0:02d}.h5'.format(i))

    if node_name.endswith('.h5'):
        node_name = node_name.replace('.h5', '')

    return node_name, fout_node


def get_node_cmd(name, tech, sam_files, res_file, out_fpath,
                 points=slice(0, 100), points_range=None,
                 sites_per_worker=None, max_workers=None,
                 logdir='./out/log_gen', output_request=('cf_mean',),
                 site_data=None, mem_util_lim=0.4, timeout=1800,
                 curtailment=None, gid_map=None, verbose=False):
    """Make a reV geneneration direct-local CLI call string.

    Parameters
    ----------
    name : str
        Name of the job to be submitted.
    tech : str
        Name of the reV technology to be analyzed.
        (e.g. pv, csp, landbasedwind, offshorewind).
    sam_files : dict | str | list
        SAM input configuration ID(s) and file path(s). Keys are the SAM
        config ID(s), top level value is the SAM path. Can also be a single
        config file str. If it's a list, it is mapped to the sorted list
        of unique configs requested by points csv.
    res_file : str
        WTK or NSRDB resource file name + path.
    out_fpath : str
        Output .h5 file path
    points : slice | str | list | tuple
        Slice/list specifying project points, string pointing to a project
    points_range : list | None
        Optional range list to run a subset of sites
    sites_per_worker : int | None
        Number of sites to be analyzed in serial on a single local core.
    max_workers : int | None
        Number of workers to use on a node. None defaults to all available
        workers.
    logdir : str
        Target directory to save log files.
    output_request : list | tuple
        Output variables requested from SAM.
    site_data : str | None
        Site-specific input data for SAM calculation. String should be a
        filepath that points to a csv, Rows match sites, columns are input
        keys. Need a "gid" column.  Input as None if no site-specific data.
    mem_util_lim : float
        Memory utilization limit (fractional).
    timeout : int | float
        Number of seconds to wait for parallel run iteration to complete
        before returning zeros. Default is 1800 seconds.
    curtailment : NoneType | str
        Pointer to a file containing curtailment input parameters or None if
        no curtailment.
    gid_map : NoneType | str
        Pointer to a gid_map csv file (can be the same as project_points) with
        gid and gid_map columns or None.
    verbose : bool
        Flag to turn on debug logging. Default is False.

    Returns
    -------
    cmd : str
        Single line command line argument to call the following CLI with
        appropriately formatted arguments based on input args:
            python -m reV.generation.cli_gen [args] direct [args] local [args]
    """

    # mark a cli arg string for main() in this module
    arg_main = '-n {}'.format(SLURM.s(name))

    # make a cli arg string for direct() in this module
    arg_direct = ['-t {}'.format(SLURM.s(tech)),
                  '-p {}'.format(SLURM.s(points)),
                  '-sf {}'.format(SLURM.s(sam_files)),
                  '-rf {}'.format(SLURM.s(res_file)),
                  '-o {}'.format(SLURM.s(out_fpath)),
                  '-spw {}'.format(SLURM.s(sites_per_worker)),
                  '-lo {}'.format(SLURM.s(logdir)),
                  '-or {}'.format(SLURM.s(output_request)),
                  '-mem {}'.format(SLURM.s(mem_util_lim))]

    if site_data:
        arg_direct.append('-sd {}'.format(SLURM.s(site_data)))

    if curtailment:
        arg_direct.append('-curt {}'.format(SLURM.s(curtailment)))

    if gid_map:
        arg_direct.append('-gm {}'.format(SLURM.s(gid_map)))

    # make a cli arg string for local() in this module
    arg_loc = ['-mw {}'.format(SLURM.s(max_workers)),
               '-to {}'.format(SLURM.s(timeout)),
               '-pr {}'.format(SLURM.s(points_range))]

    if verbose:
        arg_loc.append('-v')

    # Python command that will be executed on a node
    # command strings after cli v7.0 use dashes instead of underscores
    cmd = ('python -m reV.generation.cli_gen '
           '{arg_main} direct {arg_direct} local {arg_loc}'
           .format(arg_main=arg_main,
                   arg_direct=' '.join(arg_direct),
                   arg_loc=' '.join(arg_loc)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--nodes', '-no', default=1, type=INT,
              show_default=True,
              help='Number of SLURM nodes for gen job. Default is 1.')
@click.option('--memory', '-mem', default=None, type=INT,
              show_default=True,
              help='Single node memory request in GB. Default is None')
@click.option('--walltime', '-wt', default=1.0, type=float,
              show_default=True,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--conda_env', '-env', default=None, type=STR,
              show_default=True,
              help='Conda env to activate')
@click.option('--module', '-mod', default=None, type=STR,
              show_default=True,
              help='Module to load')
@click.option('--stdout_path', '-sout', default='./out/stdout', type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is ./out/stdout')
@click.option('--sh_script', '-sh', default=None, type=STR,
              show_default=True,
              help='Extra shell script commands to run before the reV call.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def slurm(ctx, alloc, nodes, memory, walltime, feature, conda_env, module,
          stdout_path, sh_script, verbose):
    """Run generation on HPC via SLURM job submission."""

    name = ctx.obj['NAME']
    tech = ctx.obj['TECH']
    points = ctx.obj['POINTS']
    sam_files = ctx.obj['SAM_FILES']
    res_file = ctx.obj['RES_FILE']
    sites_per_worker = ctx.obj['SITES_PER_WORKER']
    dirout, fout = os.path.split(ctx.obj['OUT_FPATH'])
    logdir = ctx.obj['LOGDIR']
    output_request = ctx.obj['OUTPUT_REQUEST']
    site_data = ctx.obj['SITE_DATA']
    max_workers = ctx.obj['MAX_WORKERS']
    mem_util_lim = ctx.obj['MEM_UTIL_LIM']
    timeout = ctx.obj['TIMEOUT']
    curtailment = ctx.obj['CURTAILMENT']
    gid_map = ctx.obj['GID_MAP']
    verbose = any([verbose, ctx.obj['VERBOSE']])

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    pc = get_node_pc(points, sam_files, tech, res_file, nodes)

    for i, split in enumerate(pc):
        node_name, fout_node = get_node_name_fout(name, fout, i, pc,
                                                  hpc='slurm')

        node_fpath = os.path.join(dirout, fout_node)
        cmd = get_node_cmd(node_name, tech, sam_files, res_file, node_fpath,
                           points=points,
                           points_range=split.split_range,
                           sites_per_worker=sites_per_worker,
                           max_workers=max_workers,
                           logdir=logdir,
                           output_request=output_request,
                           site_data=site_data,
                           mem_util_lim=mem_util_lim,
                           timeout=timeout,
                           curtailment=curtailment,
                           gid_map=gid_map,
                           verbose=verbose)

        if sh_script:
            cmd = sh_script + '\n' + cmd

        status = Status.retrieve_job_status(dirout,
                                            module=ModuleName.GENERATION,
                                            job_name=node_name,
                                            hardware='slurm',
                                            subprocess_manager=slurm_manager)

        msg = 'Generation CLI failed to submit jobs!'
        if status == 'successful':
            msg = ('Job "{}" is successful in status json found in "{}", '
                   'not re-running.'
                   .format(node_name, dirout))
        elif 'fail' not in str(status).lower() and status is not None:
            msg = ('Job "{}" was found with status "{}", not resubmitting'
                   .format(node_name, status))
        else:
            logger.info('Running reV generation on SLURM with node name "{}" '
                        'for {} (points range: {}).'
                        .format(node_name, pc, split.split_range))
            # create and submit the SLURM job
            out = slurm_manager.sbatch(cmd,
                                       alloc=alloc,
                                       memory=memory,
                                       walltime=walltime,
                                       feature=feature,
                                       name=node_name,
                                       stdout_path=stdout_path,
                                       conda_env=conda_env,
                                       module=module)[0]
            if out:
                msg = ('Kicked off reV generation job "{}" (SLURM jobid #{}).'
                       .format(node_name, out))

            # add job to reV status file.
            Status.add_job(
                dirout, module=ModuleName.GENERATION,
                job_name=node_name, replace=True,
                job_attrs={'job_id': out, 'hardware': 'slurm',
                           'fout': fout_node, 'dirout': dirout})

        click.echo(msg)
        logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Generation CLI')
        raise



# -*- coding: utf-8 -*-
"""reV scheduled losses module.

"""
import logging
import warnings
import json

import numpy as np

from reV.losses.utils import (convert_to_full_month_names,
                              filter_unknown_month_names,
                              hourly_indices_for_months)
from reV.utilities.exceptions import reVLossesValueError, reVLossesWarning


logger = logging.getLogger(__name__)


class Outage:
    """A specific type of outage.

    This class stores and validates information about a single type of
    outage. In particular, the number of outages, duration, percentage
    of farm down, and the allowed months for scheduling the outage
    must all be provided. These inputs are then validated so that they
    can be used in instances of scheduling objects.

    """

    REQUIRED_KEYS = {'count',
                     'duration',
                     'percentage_of_capacity_lost',
                     'allowed_months'}
    """Required keys in the input specification dictionary."""

    def __init__(self, specs):
        """

        Parameters
        ----------
        specs : dict
            A dictionary containing specifications for this outage. This
            dictionary must contain the following keys:
                - `count`
                    An integer value representing the total number of
                    times this outage should be scheduled. This number
                    should be larger than 0.
                - `duration`
                    An integer value representing the total number of
                    consecutive hours that this outage should take. This
                    value must be larger than 0 and less than the number
                    of hours in the allowed months.
                - `percentage_of_capacity_lost`
                    An integer or float value representing the total
                    percentage of the total capacity that will be lost
                    for the duration of the outage. This value must be
                    in the range (0, 100].
                - `allowed_months`
                    A list of month names corresponding to the allowed
                    months for the scheduled outages. Month names can be
                    unformatted and can be specified using 3-letter
                    month abbreviations.
            The input dictionary can also provide the following optional
            keys:
                - `allow_outage_overlap` - by default, ``True``
                    A bool flag indicating whether or not this outage is
                    allowed to overlap with other outages, including
                    itself. It is recommended to set this value to
                    ``True`` whenever possible, as it allows for more
                    flexible scheduling.
                - `name` - by default, string containing init parameters
                    A unique name for the outage, used for more
                    descriptive error messages.

        """
        self._specs = specs
        self._full_month_names = None
        self._total_available_hours = None
        self._name = None
        self._validate()

    def _validate(self):
        """Validate the input specs."""
        self._validate_required_keys_exist()
        self._validate_count()
        self._validate_and_convert_to_full_name_months()
        self._validate_duration()
        self._validate_percentage()

    def _validate_required_keys_exist(self):
        """Raise an error if any required keys are missing."""
        missing_keys = [n for n in self.REQUIRED_KEYS if n not in self._specs]
        if any(missing_keys):
            msg = ("The following required keys are missing from the Outage "
                   "specification: {}".format(sorted(missing_keys)))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_count(self):
        """Validate that the total number of outages is an integer. """
        if not isinstance(self.count, int):
            msg = ("Number of outages must be an integer, but got {} for {}"
                   .format(self.count, self.name))
            logger.error(msg)
            raise reVLossesValueError(msg)

        if self.count < 1:
            msg = ("Number of outages must be greater than 0, but got "
                   "{} for {}".format(self.count, self.name))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_and_convert_to_full_name_months(self):
        """Validate month input and convert to full month names. """
        months = convert_to_full_month_names(self._specs['allowed_months'])
        known_months, unknown_months = filter_unknown_month_names(months)

        if unknown_months:
            msg = ("The following month names were not understood: {}. Please "
                   "use either the full month name or the standard 3-letter "
                   "month abbreviation. For more info, see the month name "
                   "documentation for the python standard package `calendar`."
                   .format(unknown_months))
            logger.warning(msg)
            warnings.warn(msg, reVLossesWarning)

        if not known_months:
            msg = ("No known month names were provided! Please use either the "
                   "full month name or the standard 3-letter month "
                   "abbreviation. For more info, see the month name "
                   "documentation for the python standard package `calendar`. "
                   "Received input: {!r}"
                   .format(self._specs['allowed_months']))
            logger.error(msg)
            raise reVLossesValueError(msg)

        self._full_month_names = list(set(known_months))

    def _validate_duration(self):
        """Validate that the duration is between 0 and the max total. """
        if not isinstance(self.duration, int):
            msg = ("Duration must be an integer number of hours, "
                   "but got {} for {}".format(self.duration, self.name))
            logger.error(msg)
            raise reVLossesValueError(msg)

        if not 1 <= self.duration <= self.total_available_hours:
            msg = ("Duration of outage must be between 1 and the total "
                   "available hours based on allowed month input ({} for "
                   "a total hour count of {}), but got {} for {}"
                   .format(self.allowed_months, self.total_available_hours,
                           self.percentage_of_capacity_lost, self.name))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def _validate_percentage(self):
        """Validate that the percentage is in the range (0, 100]. """
        if not 0 < self.percentage_of_capacity_lost <= 100:
            msg = ("Percentage of farm down during outage must be in the "
                   "range (0, 100], but got {} for {}"
                   .format(self.percentage_of_capacity_lost, self.name))
            logger.error(msg)
            raise reVLossesValueError(msg)

    def __repr__(self):
        return "Outage({!r})".format(self._specs)

    def __str__(self):
        if self._name is None:
            self._name = self._specs.get('name') or self._default_name()
        return self._name

    def _default_name(self):
        """Generate a default name for the outage."""
        specs = self._specs.copy()
        specs.update({'allowed_months': self.allowed_months,
                      'allow_outage_overlap': self.allow_outage_overlap})
        specs_as_str = ", ".join(["{}={}".format(k, v)
                                  for k, v in specs.items()])
        return "Outage({})".format(specs_as_str)

    @property
    def count(self):
        """int: Total number of times outage should be scheduled."""
        return self._specs['count']

    @property
    def duration(self):
        """int: Total number of consecutive hours per outage."""
        return self._specs['duration']

    @property
    def percentage_of_capacity_lost(self):
        """int | float: Percent of capacity taken down per outage."""
        return self._specs['percentage_of_capacity_lost']

    @property
    def allowed_months(self):
        """list: Months during which outage can be scheduled."""
        return self._full_month_names

    @property
    def allow_outage_overlap(self):
        """bool: Indicator for overlap with other outages."""
        return self._specs.get('allow_outage_overlap', True)

    @property
    def name(self):
        """str: Name of the outage."""
        return self._specs.get('name', str(self))

    @property
    def total_available_hours(self):
        """int: Total number of hours avialbale based on allowed months."""
        if self._total_available_hours is None:
            self._total_available_hours = len(
                hourly_indices_for_months(self.allowed_months))
        return self._total_available_hours


class OutageScheduler:
    """A scheduler for multiple input outages.

    Given a list of information about different types of desired
    outages, this class leverages the stochastic scheduling routines of
    :class:`SingleOutageScheduler` to calculate the total losses due to
    the input outages on an hourly basis.

    Attributes
    ----------
    outages : :obj:`list` of :obj:`Outages <Outage>`
        The user-provided list of :obj:`Outages <Outage>` containing
        info about all types of outages to be scheduled.
    seed : :obj:`int`
        The seed value used to seed the random generator in order
        to produce random but reproducible losses. This is useful
        for ensuring that stochastically scheduled losses vary
        between different sites (i.e. that randomly scheduled
        outages in two different location do not match perfectly on
        an hourly basis).
    total_losses : :obj:`np.array`
        An array (of length 8760) containing the per-hour total loss
        percentage resulting from the stochastically scheduled outages.
        This array contains only zero values before the
        :meth:`~OutageScheduler.calculate` method is run.
    can_schedule_more : :obj:`np.array`
        A boolean array (of length 8760) indicating wether or not more
        losses can be scheduled for a given hour. This array keeps track
        of all the scheduling conflicts between input outages.

    Warnings
    --------
    It is possible that not all outages input by the user will be
    scheduled. This can happen when there is not enough time allowed
    for all of the input outages. To avoid this issue, always be sure to
    allow a large enough month range for long outages that take up a big
    portion of the farm and try to allow outage overlap whenever
    possible.

    See Also
    --------
    :class:`SingleOutageScheduler` : Single outage scheduler.
    :class:`Outage` : Specifications for a single outage.
    """

    def __init__(self, outages, seed=0):
        """
        Parameters
        ----------
        outages : list of :obj:`Outages <Outage>`
            A list of :obj:`Outages <Outage>`, where each :obj:`Outage`
            contains info about a single type of outage. See the
            documentation of :class:`Outage` for a description of the
            required keys of each outage dictionary.
        seed : int, optional
            An integer value used to seed the random generator in order
            to produce random but reproducible losses. This is useful
            for ensuring that stochastically scheduled losses vary
            between different sites (i.e. that randomly scheduled
            outages in two different location do not match perfectly on
            an hourly basis). By default, the seed is set to 0.
        """
        self.outages = outages
        self.seed = seed
        self.total_losses = np.zeros(8760)
        self.can_schedule_more = np.full(8760, True)

    def calculate(self):
        """Calculate total losses from stochastically scheduled outages.

        This function calls :meth:`SingleOutageScheduler.calculate`
        on every outage input (sorted by largest duration and then
        largest number of outages) and returns the aggregate the losses
        from the result.

        Returns
        -------
        :obj:`np.array`
            An array (of length 8760) containing the per-hour total loss
            percentage resulting from the stochastically scheduled
            outages.
        """
        sorted_outages = sorted(self.outages,
                                key=lambda o: (o.duration,
                                               o.count,
                                               o.percentage_of_capacity_lost,
                                               sum(sum(map(ord, name))
                                                   for name
                                                   in o.allowed_months),
                                               o.allow_outage_overlap))
        for outage in sorted_outages[::-1]:
            self.seed += 1
            SingleOutageScheduler(outage, self).calculate()
        return self.total_losses


class SingleOutageScheduler:
    """A scheduler for a single outage.

    Given information about a single type of outage, this class
    facilitates the (randomized) scheduling of all requested instances
    of the outage. See :meth:`SingleOutageScheduler.calculate` for
    specific details about the scheduling process.

    Attributes
    ----------
    outage : :obj:`Outage`
        The user-provided :obj:`Outage` containing info about the outage
        to be scheduled.
    scheduler : :obj:`OutageScheduler`
        A scheduler object that keeps track of the total hourly losses
        from the input outage as well as any other outages it has
        already scheduled.
    can_schedule_more : :obj:`np.array`
        A boolean array (of length 8760) indicating wether or not more
        losses can be scheduled for a given hour. This is specific
        to the input outage only.

    Warnings
    --------
    It is possible that not all outages input by the user can be
    scheduled. This can happen when there is not enough time allowed
    for all of the input outages. To avoid this issue, always be sure to
    allow a large enough month range for long outages that take up a big
    portion of the farm and try to allow outage overlap whenever
    possible.

    See Also
    --------
    :class:`OutageScheduler` : Scheduler for multiple outages.
    :class:`Outage` : Specifications for a single outage.
    """

    MAX_ITER = 10_000
    """Max number of extra attempts to schedule outages."""

    def __init__(self, outage, scheduler):
        """

        Parameters
        ----------
        outage : Outage
            An outage object contianing info about the outage to be
            scheduled.
        scheduler : OutageScheduler
            A scheduler object that keeps track of the total hourly
            losses from the input outage as well as any other outages
            it has already scheduled.
        """
        self.outage = outage
        self.scheduler = scheduler
        self.can_schedule_more = np.full(8760, False)
        self._scheduled_outage_inds = []

    def calculate(self):
        """Calculate losses from stochastically scheduled outages.

        This function attempts to schedule outages according to the
        specification provided in the :obj:`Outage` input. Specifically,
        it checks the available hours based on the main
        :obj:`Scheduler <OutageScheduler>` (which may have other outages
        already scheduled) and attempts to randomly add new outages with
        the specified duration and percent of losses. The function
        terminates when the desired number of outages (specified by
        :attr:`Outage.count`) have been successfully scheduled, or when
        the number of attempts exceeds
        :attr:`~SingleOutageScheduler.MAX_ITER` + :attr:`Outage.count`.

        Warns
        -----
        reVLossesWarning
            If the number of requested outages could not be scheduled.
        """
        self.update_when_can_schedule_from_months()

        for iter_ind in range(self.outage.count + self.MAX_ITER):
            self.update_when_can_schedule()
            if not self.can_schedule_more.any():
                break
            seed = self.scheduler.seed + iter_ind
            outage_slice = self.find_random_outage_slice(seed=seed)
            if self.can_schedule_more[outage_slice].all():
                self.schedule_losses(outage_slice)
            if len(self._scheduled_outage_inds) == self.outage.count:
                break

        if len(self._scheduled_outage_inds) < self.outage.count:
            if len(self._scheduled_outage_inds) == 0:
                msg_start = "Could not schedule any requested outages"
            else:
                msg_start = ("Could only schedule {} out of {} requested "
                             "outages"
                             .format(len(self._scheduled_outage_inds),
                                     self.outage.count))
            msg = ("{} after a max of {:,} iterations. You are likely "
                   "attempting to schedule a lot of long outages or a lot "
                   "of short outages with a large percentage of the farm at "
                   "a time. Please adjust the outage specifications and try "
                   "again"
                   .format(msg_start, self.outage.count + self.MAX_ITER))
            logger.warning(msg)
            warnings.warn(msg, reVLossesWarning)

    def update_when_can_schedule_from_months(self):
        """
        Update :attr:`can_schedule_more` using :attr:`Outage.allowed_months`.

        This function sets the :attr:`can_schedule_more` bool array to
        `True` for all of the months in :attr:`Outage.allowed_months`.
        """
        inds = hourly_indices_for_months(self.outage.allowed_months)
        self.can_schedule_more[inds] = True

    def update_when_can_schedule(self):
        """Update :attr:`can_schedule_more` using :obj:`OutageScheduler`.

        This function sets the :attr:`can_schedule_more` bool array to
        `True` wherever :attr:`OutageScheduler.can_schedule_more` is
        also `True` and wherever the losses from this outage would not
        cause the :attr:`OutageScheduler.total_losses` to exceed 100%.
        """
        self.can_schedule_more &= self.scheduler.can_schedule_more
        if self.outage.allow_outage_overlap:
            total_new_losses = (self.scheduler.total_losses
                                + self.outage.percentage_of_capacity_lost)
            losses_will_not_exceed_100 = total_new_losses <= 100
            self.can_schedule_more &= losses_will_not_exceed_100
        else:
            self.can_schedule_more &= self.scheduler.total_losses == 0

    def find_random_outage_slice(self, seed=None):
        """Find a random slot of time for this type of outage.

        This function randomly selects a starting time for this outage
        given the allowed times in :attr:`can_schedule_more`. It does
        **not** verify that the outage can be scheduled for the entire
        requested duration.

        Parameters
        ----------
        seed : int, optional
            Integer used to seed the :func:`np.random.choice` call.
            If :obj:`None`, seed is not used.

        Returns
        -------
        :obj:`slice`
            A slice corresponding to the random slot of time for this
            type of outage.
        """
        if seed is not None:
            np.random.seed(seed)
        outage_ind = np.random.choice(np.where(self.can_schedule_more)[0])
        return slice(outage_ind, outage_ind + self.outage.duration)

    def schedule_losses(self, outage_slice):
        """Schedule the input outage during the given slice of time.

        Given a slice in the hourly loss array, add the losses from this
        outage (which is equivalent to scheduling them).

        Parameters
        ----------
        outage_slice : slice
            A slice corresponding to the slot of time to schedule this
            outage.
        """
        self._scheduled_outage_inds.append(outage_slice.start)
        self.scheduler.total_losses[outage_slice] += (
            self.outage.percentage_of_capacity_lost)
        if not self.outage.allow_outage_overlap:
            self.scheduler.can_schedule_more[outage_slice] = False


class ScheduledLossesMixin:
    """Mixin class for :class:`reV.SAM.generation.AbstractSamGeneration`.

    Warning
    -------
    Using this class for anything excpet as a mixin for
    :class:`~reV.SAM.generation.AbstractSamGeneration` may result in
    unexpected results and/or errors.
    """

    OUTAGE_CONFIG_KEY = 'reV_outages'
    """Specify outage information in the config file using this key."""
    OUTAGE_SEED_CONFIG_KEY = 'reV_outages_seed'
    """Specify a randomizer seed in the config file using this key."""

    def add_scheduled_losses(self, resource):
        """Add stochastically scheduled losses to SAM config file.

        This function reads the information in the ``reV_outages`` key
        of the ``sam_sys_inputs`` dictionary and computes stochastically
        scheduled losses from that input. If the value for
        ``reV_outages`` is a string, it must have been generated by
        calling :func:`json.dumps` on the list of dictionaries
        containing outage specifications. Otherwise, the outage
        information is expected to be a list of dictionaries containing
        outage specifications. See :class:`Outage` for a description of
        the specifications allowed for each outage. The scheduled losses
        are passed to SAM via the ``hourly`` key to signify which hourly
        capacity factors should be adjusted with outage losses. If no
        outage info is specified in ``sam_sys_inputs``, no scheduled
        losses are added.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries resource data for a single location with a pandas
            DatetimeIndex. The ``year`` value of the index will be used
            to seed the stochastically scheduled losses.

        See Also
        --------
        :class:`Outage` : Single outage specification.

        Notes
        -----
        The scheduled losses are passed to SAM via the ``hourly`` key to
        signify which hourly capacity factors should be adjusted with
        outage losses. If the user specifies other hourly adjustment
        factors via the ``hourly`` key, the effect is combined. For
        example, if the user inputs a 33% hourly adjustment factor and
        reV schedules an outage for 70% of the farm down for the same
        hour, then the resulting adjustment factor is::
            1 - [(1 - 70/100) * (1 - 33/100)] = 0.799
        This means the generation will be reduced by ~80%, because the
        user requested 33% losses for the 30% the farm that remained
        operational during the scheduled outage (i.e. 20% remaining of
        the original generation).
        """
        outages = self._user_outage_input(resource)
        if not outages:
            return

        logger.debug("Adding the following stochastically scheduled outages: "
                     "{}".format(outages))
        logger.debug("Scheduled outages seed: {}".format(self.outage_seed))

        scheduler = OutageScheduler(outages, seed=self.outage_seed)
        hourly_outages = scheduler.calculate()
        self._add_outages_to_sam_inputs(hourly_outages)

        logger.debug("Hourly adjustment factors after scheduled outages: {}"
                     .format(list(self.sam_sys_inputs['hourly'])))

    def _user_outage_input(self, resource):
        """Get outage and seed info from config. """
        outage_specs = self.sam_sys_inputs.pop(self.OUTAGE_CONFIG_KEY, None)
        if outage_specs is None:
            return

        # site-specific info is input as str
        if isinstance(outage_specs, str):
            outage_specs = json.loads(outage_specs)

        outages = [Outage(spec) for spec in outage_specs]
        self.__year_seed = int(resource.index.year.values[0])
        self.__user_input_seed = self.sam_sys_inputs.pop(
            self.OUTAGE_SEED_CONFIG_KEY, 0)
        return outages

    def _add_outages_to_sam_inputs(self, outages):
        """Add the hourly adjustment factors to config, checking user input."""

        hourly_mult = 1 - outages / 100

        user_hourly_input = self.sam_sys_inputs.pop('hourly', [0] * 8760)
        user_hourly_mult = 1 - np.array(user_hourly_input) / 100

        final_hourly_mult = hourly_mult * user_hourly_mult
        self.sam_sys_inputs['hourly'] = (1 - final_hourly_mult) * 100

    @property
    def outage_seed(self):
        """int: A value to use as the seed for the outage losses. """
        base_seed = self.__year_seed + self.__user_input_seed
        try:
            return int(self.meta.name) + base_seed
        except (AttributeError, TypeError, ValueError):
            pass

        try:
            return hash(tuple(self.meta)) + base_seed
        except (AttributeError, TypeError):
            pass

        return base_seed



# -*- coding: utf-8 -*-
"""
functions to plot turbine layouts and boundary polygons
"""
import numpy as np
import matplotlib.pyplot as plt


def get_xy(A):
    """separate polygon exterior coordinates to x and y
    Parameters
    ----------
    A : Polygon.exteroir.coords
    exterior coordinates from a shapely Polygon

    Outputs
    ----------
    x : array
    boundary polygon x coordinates
    y : array
    boundary polygon y coordinates
    """
    x = np.zeros(len(A))
    y = np.zeros(len(A))
    for i, _ in enumerate(A):
        x[i] = A[i][0]
        y[i] = A[i][1]
    return x, y


def plot_poly(geom, ax=None, color="black", linestyle="--", linewidth=0.5):
    """plot the wind plant boundaries
    Parameters
    ----------
    geom : Polygon | MultiPolygon
        the shapely.Polygon or shapely.MultiPolygon that define the wind plant
    boundary(ies).
    ax (:py:class:`matplotlib.pyplot.axes`, optional):
        The figure axes on which the wind rose is plotted. Defaults to None.
    color : string (optional)
        the color for the wind plant boundaries
    linestyle : string (optional)
        style to plot the boundary lines
    linewidth : float (optional)
        the width of the boundary lines
    """
    if ax is None:
        _, ax = plt.subplots()

    if geom.type == 'Polygon':
        exterior_coords = geom.exterior.coords[:]
        x, y = get_xy(exterior_coords)
        ax.fill(x, y, color="C0", alpha=0.25)
        ax.plot(x, y, color=color, linestyle=linestyle, linewidth=linewidth)

        for interior in geom.interiors:
            interior_coords = interior.coords[:]
            x, y = get_xy(interior_coords)
            ax.fill(x, y, color="white", alpha=1.0)
            ax.plot(x, y, "--k", linewidth=0.5)

    elif geom.type == 'MultiPolygon':

        for part in geom:
            exterior_coords = part.exterior.coords[:]
            x, y = get_xy(exterior_coords)
            ax.fill(x, y, color="C0", alpha=0.25)
            ax.plot(x, y, color=color, linestyle=linestyle,
                    linewidth=linewidth)

            for interior in part.interiors:
                interior_coords = interior.coords[:]
                x, y = get_xy(interior_coords)
                ax.fill(x, y, color="white", alpha=1.0)
                ax.plot(x, y, "--k", linewidth=0.5)
    return ax


def plot_turbines(x, y, r, ax=None, color="C0", nums=False):
    """plot wind turbine locations
    Parameters
    ----------
    x : array
        wind turbine x locations
    y : array
        wind turbine y locations
    r : float
        wind turbine radius
    ax (:py:class:`matplotlib.pyplot.axes`, optional):
        The figure axes on which the wind rose is plotted. Defaults to None.
    color : string (optional)
        the color for the wind plant boundaries
    nums : Bool (optional)
        show the turbine numbers next to each turbine
    """
    # Set up figure
    if ax is None:
        _, ax = plt.subplots()

    n = len(x)
    for i in range(n):
        t = plt.Circle((x[i], y[i]), r, color=color)
        ax.add_patch(t)
        if nums is True:
            ax.text(x[i], y[i], "%s" % (i + 1))

    return ax


def plot_windrose(wind_directions, wind_speeds, wind_frequencies, ax=None,
                  colors=None):
    """plot windrose
    Parameters
    ----------
    wind_directions : 1D array
        wind direction samples
    wind_speeds : 1D array
        wind speed samples
    wind_frequencies : 2D array
        frequency of wind direction and speed samples
    ax (:py:class:`matplotlib.pyplot.axes`, optional):
        The figure axes on which the wind rose is plotted. Defaults to None.
    color : array (optional)
        the color for the different wind speed bins
    """
    if ax is None:
        _, ax = plt.subplots(subplot_kw=dict(polar=True))

    ndirs = len(wind_directions)
    nspeeds = len(wind_speeds)

    if colors is None:
        colors = []
        for i in range(nspeeds):
            colors = np.append(colors, "C%s" % i)

    for i in range(ndirs):
        wind_directions[i] = np.deg2rad(90.0 - wind_directions[i])

    width = 0.8 * 2 * np.pi / len(wind_directions)

    for i in range(ndirs):
        bottom = 0.0
        for j in range(nspeeds):
            if i == 0:
                if j < nspeeds - 1:
                    ax.bar(wind_directions[i], wind_frequencies[j, i],
                           bottom=bottom, width=width, edgecolor="black",
                           color=[colors[j]],
                           label="%s-%s m/s" % (int(wind_speeds[j]),
                                                int(wind_speeds[j + 1]))
                           )
                else:
                    ax.bar(wind_directions[i], wind_frequencies[j, i],
                           bottom=bottom, width=width, edgecolor="black",
                           color=[colors[j]],
                           label="%s+ m/s" % int(wind_speeds[j])
                           )
            else:
                ax.bar(wind_directions[i], wind_frequencies[j, i],
                       bottom=bottom, width=width, edgecolor="black",
                       color=[colors[j]])
            bottom = bottom + wind_frequencies[j, i]

    ax.legend(bbox_to_anchor=(1.3, 1), fontsize=10)
    pi = np.pi
    ax.set_xticks((0, pi / 4, pi / 2, 3 * pi / 4, pi, 5 * pi / 4,
                   3 * pi / 2, 7 * pi / 4))
    ax.set_xticklabels(("E", "NE", "N", "NW", "W", "SW", "S", "SE"),
                       fontsize=10)
    plt.yticks(fontsize=10)

    plt.subplots_adjust(left=0.0, right=1.0, top=0.9, bottom=0.1)

    return ax



# -*- coding: utf-8 -*-
"""
reV bespoke wind plant optimization config

Created on Jan 2022

@author: gbuster
"""
import logging

from reV.config.output_request import SAMOutputRequest
from reV.config.base_analysis_config import AnalysisConfig
from reV.utilities.exceptions import ConfigError
from reV.utilities import ModuleName

logger = logging.getLogger(__name__)


class BespokeConfig(AnalysisConfig):
    """SAM-based analysis config (generation, lcoe, etc...)."""

    NAME = ModuleName.BESPOKE

    REQUIREMENTS = ('excl_fpath', 'res_fpath', 'tm_dset', 'objective_function',
                    'capital_cost_function', 'fixed_operating_cost_function',
                    'variable_operating_cost_function',
                    'project_points', 'sam_files',
                    )

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            File path to config json (str), serialized json object (str),
            or dictionary with pre-extracted config.
        """
        super().__init__(config)

    def _preflight(self):
        """Run preflight checks for missing REQUIREMENTS and also check special
        bespoke inputs."""
        super()._preflight()

        if self.project_points is None and len(self.sam_files) > 1:
            msg = ('If project_points is None, only one sam_files entry '
                   'should be present, but received {} sam_files: {}'
                   .format(len(self.sam_files), self.sam_files))
            logger.error(msg)
            raise ConfigError(msg)

    @property
    def excl_fpath(self):
        """Get the exclusions filepath(s), Required."""
        return self['excl_fpath']

    @property
    def res_fpath(self):
        """
        Wind resource h5 filepath in NREL WTK format. Can also include
        unix-style wildcards like /dir/wind_*.h5 for multiple years of
        resource data. Required.

        Returns
        -------
        str
        """
        return self['res_fpath']

    @property
    def tm_dset(self):
        """Get the techmap dataset, required."""
        return self['tm_dset']

    @property
    def objective_function(self):
        """The objective function of the optimization as a string

        Should return the objective to be minimized during layout optimization.
        Variables available are:
            - n_turbines: the number of turbines
            - system_capacity: wind plant capacity
            - aep: annual energy production
            - fixed_charge_rate: user input fixed_charge_rate if included
              as part of the sam system config.
            - self.wind_plant: the SAM wind plant object, through which
            all SAM variables can be accessed
            - capital_cost: plant capital cost as evaluated
              by `capital_cost_function`
            - fixed_operating_cost: plant fixed annual operating cost as
              evaluated by `fixed_operating_cost_function`
            - variable_operating_cost: plant variable annual operating cost
              as evaluated by `variable_operating_cost_function`
        """
        return self['objective_function']

    @property
    def capital_cost_function(self):
        """The plant capital cost function as a string, must return the total
        capital cost in $. Has access to the same variables as the
        objective_function."""
        return self['capital_cost_function']

    @property
    def fixed_operating_cost_function(self):
        """The plant annual fixed operating cost function as a string, must
        return the fixed operating cost in $/year. Has access to the same
        variables as the objective_function."""
        return self['fixed_operating_cost_function']

    @property
    def variable_operating_cost_function(self):
        """The plant annual variable operating cost function as a string, must
        return the variable operating cost in $/kWh. Has access to the same
        variables as the objective_function."""
        return self['variable_operating_cost_function']

    @property
    def project_points(self):
        """This can be None to use all available reV supply curve points or a
        string pointing to a project points csv. Points csv should have 'gid'
        and 'config' column, the config maps to the sam_configs dict keys.

        Returns
        -------
        pp : ProjectPoints
            ProjectPoints object
        """
        return self['project_points']

    @property
    def points_range(self):
        """An optional input that specifies the (start, end) index (inclusive,
        exclusive) of the project points to analyze. If this is specified, the
        requested points are analyzed on a single worker.
        """
        return self.get('points_range', None)

    @property
    def sam_files(self):
        """SAM config files. This should be a dictionary mapping config ids
        (keys) to config filepaths. If points is None, only one entry should be
        present and can be a single string filepath (no dict necessary).

        Returns
        -------
        dict
        """
        sf = self['sam_files']
        if isinstance(sf, str):
            sf = {'default': sf}
        return sf

    @property
    def min_spacing(self):
        """Minimum spacing between turbines in meters. Can also be a string
        like "5x" (default) which is interpreted as 5 times the turbine rotor
        diameter.
        """
        return str(self.get('min_spacing', '5x'))

    @property
    def wake_loss_multiplier(self):
        """Wake loss multiplier. Default is 1. Only applied during
        optimization."""
        return self.get('wake_loss_multiplier', 1)

    @property
    def ga_kwargs(self):
        """Get the ga_kwargs dictionary. Default is None (default init)."""
        return self.get('ga_kwargs', None)

    @property
    def output_request(self):
        """Get the list of requested output variables. default is
        ('system_capacity', 'cf_mean')
        """

        default = ('system_capacity', 'cf_mean')
        _output_request = self.get('output_request', default)
        _output_request = SAMOutputRequest(_output_request)

        return _output_request

    @property
    def ws_bins(self):
        """Get the windspeed binning arguments This should be a 3-entry list
        with [start, stop, step] for the windspeed binning of the wind joint
        probability distribution. The stop value is inclusive, so ws_bins=[0,
        20, 5] would result in four bins with bin edges [0, 5, 10, 15, 20].
        Default is [0, 20, 5].
        """
        return self.get('ws_bins', [0, 20, 5])

    @property
    def wd_bins(self):
        """Get the winddirection binning arguments This should be a 3-entry
        list with [start, stop, step] for the winddirection binning of the
        wind joint probability distribution. The stop value is inclusive, so
        ws_bins=[0, 360, 90] would result in four bins with bin edges [0, 90,
        180, 270, 360]. Default is [0, 360, 45].
        """
        return self.get('wd_bins', [0, 360, 45])

    @property
    def excl_dict(self):
        """Get the exclusions dictionary. Default is None (all included)."""
        return self.get('excl_dict', None)

    @property
    def area_filter_kernel(self):
        """Get the minimum area filter kernel name ('queen' or 'rook'). Default
        is queen but the area filter kernel isnt active unless min_area != 0"""
        return self.get('area_filter_kernel', 'queen')

    @property
    def min_area(self):
        """Get the minimum area filter minimum area in km2. Default is None
        (not active)."""
        return self.get('min_area', None)

    @property
    def resolution(self):
        """Get the SC resolution. Default is 64."""
        return self.get('resolution', 64)

    @property
    def excl_area(self):
        """Get the exclusion pixel area in km2. Default is None which will
        determine the area from the exclusion file projection profile."""
        return self.get('excl_area', None)

    @property
    def data_layers(self):
        """Aggregation data layers. Must be a dictionary keyed by data label
        name. Each value must be another dictionary with "dset", "method",
        and "fpath".
        """
        return self.get('data_layers', None)

    @property
    def pre_extract_inclusions(self):
        """Optional flag to pre-extract/compute the inclusion mask from the
        provided excl_dict, by default False. Typically faster to compute
        the inclusion mask on the fly with parallel workers.
        """
        return self.get('pre_extract_inclusions', False)



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Multi-year means CLI entry points.
"""
import click
import json
import logging
import os
import time

from reV.config.multi_year import MultiYearConfig
from reV.handlers.multi_year import MultiYear
from reV.pipeline.status import Status
from reV.utilities import ModuleName
from reV import __version__

from rex.utilities.cli_dtypes import STR, STRLIST, PATHLIST, INT
from rex.utilities.loggers import init_mult
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=os.path.basename(os.getcwd()),
              type=STR, show_default=True, help='reV Multi-Year job name.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Multi-Year Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid MultiYear config keys
    """
    click.echo(', '.join(get_class_properties(MultiYearConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV multi-year configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV gen from a config file."""

    # Instantiate the config object
    config = MultiYearConfig(config_file)

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # make output directory if does not exist
    if not os.path.exists(config.dirout):
        os.makedirs(config.dirout)

    # initialize loggers.
    init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV multi-year from config file: "{}"'
                .format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))

    ctx.obj['MY_FILE'] = config.my_file
    if config.execution_control.option == 'local':

        ctx.obj['NAME'] = name
        status = Status.retrieve_job_status(
            config.dirout, module=ModuleName.MULTI_YEAR, job_name=name
        )
        if status != 'successful':
            Status.add_job(
                config.dirout, module=ModuleName.MULTI_YEAR,
                job_name=name, replace=True,
                job_attrs={'hardware': 'local',
                           'fout': ctx.obj['MY_FILE'],
                           'dirout': config.dirout})
            group_params = json.dumps(config.group_params)
            ctx.invoke(multi_year_groups, group_params=group_params)

    elif config.execution_control.option in ('eagle', 'slurm'):
        ctx.obj['NAME'] = name
        ctx.invoke(multi_year_slurm,
                   alloc=config.execution_control.allocation,
                   walltime=config.execution_control.walltime,
                   feature=config.execution_control.feature,
                   memory=config.execution_control.memory,
                   conda_env=config.execution_control.conda_env,
                   module=config.execution_control.module,
                   stdout_path=os.path.join(config.log_directory, 'stdout'),
                   group_params=json.dumps(config.group_params),
                   sh_script=config.execution_control.sh_script,
                   verbose=verbose)


@main.group()
@click.option('--my_file', '-f', required=True, type=click.Path(),
              help='h5 file to use for multi-year collection.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def direct(ctx, my_file, verbose):
    """Main entry point for collection with context passing."""
    ctx.obj['MY_FILE'] = my_file
    ctx.obj['VERBOSE'] = verbose


@direct.command()
@click.option('--source_files', '-sf', required=True, type=PATHLIST,
              help='List of files to collect from. Either this or '
              'source_pattern must be specified.')
@click.option('--source_pattern', '-sp', required=False,
              type=str, default=None,
              help='Optional argument to specify a unix-style '
              '/filepath/pattern*.h5 to get source files. Either this or '
              'source_files must be specified.')
@click.option('--group', '-g', type=STR, default=None,
              show_default=True,
              help='Group to collect into. Useful for collecting multiple '
              'scenarios into a single file. Optional.')
@click.option('--dsets', '-ds', required=True, type=STRLIST,
              help='Dataset names to be collected. If means, multi-year '
              'means will be computed.')
@click.option('--pass_through_dsets', '-pt', default=None, type=STRLIST,
              show_default=True,
              help='Optional list of datasets that are identical in the '
              'multi-year files (e.g. input datasets that dont vary '
              'from year to year) that should be copied to the output '
              'multi-year file once without a year suffix or '
              'means/stdev calculation')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging.')
@click.pass_context
def multi_year(ctx, source_files, source_pattern, group, dsets,
               pass_through_dsets, verbose):
    """Run multi year collection and means on local worker."""

    name = ctx.obj['NAME']
    my_file = ctx.obj['MY_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])

    # initialize loggers for multiple modules
    log_dir = os.path.dirname(my_file)
    init_mult(name, log_dir, modules=[__name__, 'reV.handlers.multi_year'],
              verbose=verbose, node=True)

    for key, val in ctx.obj.items():
        logger.debug('ctx var passed to collection method: "{}" : "{}" '
                     'with type "{}"'.format(key, val, type(val)))

    if source_pattern is not None:
        source_files = source_pattern

    logger.info('Multi-year collection is being run for "{}" '
                'with job name "{}" on {}. Target output path is: {}'
                .format(dsets, name, source_files, my_file))
    t0 = time.time()

    for dset in dsets:
        if MultiYear.is_profile(source_files, dset):
            MultiYear.collect_profiles(my_file, source_files, dset,
                                       group=group)
        else:
            MultiYear.collect_means(my_file, source_files, dset,
                                    group=group)

    if pass_through_dsets is not None:
        for dset in pass_through_dsets:
            MultiYear.pass_through(my_file, source_files, dset,
                                   group=group)

    runtime = (time.time() - t0) / 60
    logger.info('Multi-year collection completed in: {:.2f} min.'
                .format(runtime))

    # add job to reV status file.
    status = {'dirout': os.path.dirname(my_file),
              'fout': os.path.basename(my_file),
              'job_status': 'successful',
              'runtime': runtime,
              'finput': source_files}
    Status.make_job_file(os.path.dirname(my_file), ModuleName.MULTI_YEAR, name,
                         status)


@direct.command()
@click.option('--group_params', '-gp', required=True, type=str,
              help='Stringified dictionary of collection groups and their '
              'parameters, e.g.: '
              '{group1: {group: null, source_files: [], dsets: []}}')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def multi_year_groups(ctx, group_params, verbose):
    """Run multi year collection and means for multiple groups."""
    name = ctx.obj['NAME']
    my_file = ctx.obj['MY_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])

    # initialize loggers for multiple modules
    log_dir = os.path.dirname(my_file)
    init_mult(name, log_dir, modules=[__name__, 'reV.handlers.multi_year'],
              verbose=verbose, node=True)

    for key, val in ctx.obj.items():
        logger.debug('ctx var passed to collection method: "{}" : "{}" '
                     'with type "{}"'.format(key, val, type(val)))

    logger.info('Multi-year collection is being run with job name "{}". '
                'Target output path is: {}'
                .format(name, my_file))
    ts = time.time()

    for group_name, group in json.loads(group_params).items():
        logger.info('- Collecting datasets "{}" from "{}" into "{}/"'
                    .format(group['dsets'], group['source_files'],
                            group_name))
        t0 = time.time()
        for dset in group['dsets']:
            if MultiYear.is_profile(group['source_files'], dset):
                MultiYear.collect_profiles(my_file, group['source_files'],
                                           dset, group=group['group'])
            else:
                MultiYear.collect_means(my_file, group['source_files'],
                                        dset, group=group['group'])

        if group.get('pass_through_dsets', None) is not None:
            for dset in group['pass_through_dsets']:
                MultiYear.pass_through(my_file, group['source_files'],
                                       dset, group=group['group'])

        runtime = (time.time() - t0) / 60
        logger.info('- {} collection completed in: {:.2f} min.'
                    .format(group_name, runtime))

    runtime = (time.time() - ts) / 60
    logger.info('Multi-year collection completed in : {:.2f} min.'
                .format(runtime))

    # add job to reV status file.
    status = {'dirout': os.path.dirname(my_file),
              'fout': os.path.basename(my_file),
              'job_status': 'successful',
              'runtime': runtime}
    Status.make_job_file(os.path.dirname(my_file), ModuleName.MULTI_YEAR, name,
                         status)


def get_slurm_cmd(name, my_file, group_params, verbose=False):
    """Make a reV multi-year collection local CLI call string.

    Parameters
    ----------
    my_file : str
        Path to .h5 file to use for multi-year collection.
    group_params : list
        List of groups and their parameters to collect
    verbose : bool
        Flag to turn on DEBUG logging

    Returns
    -------
    cmd : str
        Argument to call the neccesary CLI calls on the node to collect
        desired groups
    """
    # make a cli arg string for direct() in this module
    main_args = ['-n {}'.format(SLURM.s(name))]

    if verbose:
        main_args.append('-v')

    direct_args = '-f {}'.format(SLURM.s(my_file))

    collect_args = '-gp {}'.format(SLURM.s(group_params))

    # Python command that will be executed on a node
    # command strings after cli v7.0 use dashes instead of underscores
    cmd = ('python -m reV.handlers.cli_multi_year {} direct {} '
           'multi-year-groups {}'
           .format(' '.join(main_args), direct_args, collect_args))
    logger.debug('Creating the following command line call:\n\t{}'
                 .format(cmd))

    return cmd


@direct.command()
@click.option('--group_params', '-gp', required=True, type=str,
              help='Stringified dictionary of collection groups and their '
              'parameters, e.g.: '
              '{group1: {group: null, source_files: [], dsets: []}}')
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--walltime', '-wt', default=4.0, type=float,
              show_default=True,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help='Additional flags for SLURM job. Format is "--qos=high" '
              'or "--depend=[state:job_id]". Default is None.')
@click.option('--memory', '-mem', default=None, type=INT,
              show_default=True,
              help='SLURM node memory request in GB. Default is None')
@click.option('--conda_env', '-env', default=None, type=STR,
              show_default=True,
              help='Conda env to activate')
@click.option('--module', '-mod', default=None, type=STR, show_default=True,
              help='Module to load')
@click.option('--stdout_path', '-sout', default='./out/stdout', type=str,
              show_default=True,
              help='Subprocess standard output path. Default is ./out/stdout')
@click.option('--sh_script', '-sh', default=None, type=STR,
              show_default=True,
              help='Extra shell script commands to run before the reV call.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def multi_year_slurm(ctx, group_params, alloc, walltime, feature, memory,
                     conda_env, module, stdout_path, sh_script, verbose):
    """
    Run multi year collection and means on HPC via SLURM job submission.
    """

    name = ctx.obj['NAME']
    my_file = ctx.obj['MY_FILE']
    verbose = any([verbose, ctx.obj['VERBOSE']])

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(os.path.dirname(my_file),
                                        module=ModuleName.MULTI_YEAR,
                                        job_name=name, hardware='slurm',
                                        subprocess_manager=slurm_manager)

    msg = 'Multi-year CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, os.path.dirname(my_file)))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV multi-year collection on SLURM with node '
                    ' name "{}", collecting into "{}".'
                    .format(name, my_file))
        # create and submit the SLURM job
        cmd = get_slurm_cmd(name, my_file, group_params, verbose=verbose)
        if sh_script:
            cmd = sh_script + '\n' + cmd
        out = slurm_manager.sbatch(cmd, alloc=alloc, memory=memory,
                                   walltime=walltime, feature=feature,
                                   name=name, stdout_path=stdout_path,
                                   conda_env=conda_env, module=module)[0]
        if out:
            msg = ('Kicked off reV multi-year collection job "{}" '
                   '(SLURM jobid #{}).'.format(name, out))

        # add job to reV status file.
        Status.add_job(
            os.path.dirname(my_file), module=ModuleName.MULTI_YEAR,
            job_name=name, replace=True,
            job_attrs={'job_id': out, 'hardware': 'slurm',
                       'fout': os.path.basename(my_file),
                       'dirout': os.path.dirname(my_file)})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Multi-Year CLI')
        raise
